/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 * 
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.pc.util;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;

/*
 * Usage of XSharedMemory:
 * 
 * DOES NOT WORK RELIABLY: MAPPED FILES DO NOT WORK PROPERLY
 * AS MEMORY REPRESENTATION (TESTED ON LINUX). ON WINDOWS, MAPPED FILES ARE
 * EXPECTED TO WORK SLOW (NO IN-MEMORY FILES PROVIDED BY OPERATING SYSTEM).
 * PROVIDE OWN NATIVE IMPLEMENTATION FOR SHARED MEMORY REPRESENTATION?
 * 
 * 
 * There must be exactly one "master" process controlling the life cycle
 * of the shared memory in the backend (see constructor for XSharedMemory;
 * the backend typically is a temporary file being mapped into memory).
 * The master process should be the first to create the shared memory
 * instance and the last to destroy it (typically on exit).
 * After the shared memory object has been destroyed by the master process,
 * there must not be any further read or write access by other processes
 * (the executing threads will probably hang).
 * 
 * Also, there must be only one writing process to a XSharedMemory
 * instance which is not necessarily identical to the master process.
 * There can be several reading processes at the same time, though.
 */
public class XSharedMemory
{
    private static final boolean WINDOWS = ('\\' == File.separatorChar);
 

    abstract class BMemoryRepresentation
    {
        abstract void shutdown() throws Throwable;
        abstract ByteBuffer getBuffer();
        abstract void read();
        abstract void write();

        @Override
        protected void finalize() throws Throwable
        {
            shutdown();
            super.finalize();
        }
    }

    private class XMappedFile extends BMemoryRepresentation
    {
        private RandomAccessFile mFile;
        private FileChannel mFileChannel;
        private MappedByteBuffer mMappedBuffer;

        XMappedFile(String name, int byte_size)
        {
            super();

            try
            {
                File file = new File(name);
                if (mMaster) file.deleteOnExit();
                mFile = new RandomAccessFile(file, "rw");
                mFileChannel = mFile.getChannel();
                mMappedBuffer = mFileChannel.map(FileChannel.MapMode.READ_WRITE, 0, byte_size);
                mMappedBuffer.limit(byte_size);
            }
            catch (SecurityException | IOException e)
            {
                System.err.println(e.getMessage());
            }
        }

        XMappedFile(File file, int byte_size)
        {
            try
            {
                if (mMaster) file.deleteOnExit();
                mFile = new RandomAccessFile(file, "rw");
                mFileChannel = mFile.getChannel();
                mMappedBuffer = mFileChannel.map(FileChannel.MapMode.READ_WRITE, 0, byte_size);
                mMappedBuffer.limit(byte_size);
            }
            catch (SecurityException | IOException e)
            {
                System.err.println(e.getMessage());
            }
        }

        @Override
        ByteBuffer getBuffer()
        {
            return mMappedBuffer;
        }

        @Override
        void read()
        {
            mMappedBuffer.load();
        }

        @Override
        void write()
        {
            mMappedBuffer.force();
        }

        @Override
        void shutdown() throws Throwable
        {
            if (null != mFileChannel)
            {
                try
                {
                    mFileChannel.close();
                }
                catch (IOException e) {}
            }

            if (null != mFile)
            {
                try
                {
                    mFile.close();
                }
                catch (IOException e) {}
            }
        }
    }



    private final boolean mMaster;
    private BMemoryRepresentation mMemory;

    public XSharedMemory(String name, int byte_size, boolean is_master)
    {
        /*
         * Windows:
         * Use temporary file mapped into memory for now.
         * Eventually provide new class derived form BMemoryRepresention
         * that allocates shared memory with native code and wrap byte array
         * by ByteBuffer (see ByteBuffer.wrap)
         * => take care of resources in shutdown()
         */
        mMaster = is_master;
        try
        {
            mMemory = WINDOWS ? new XMappedFile(File.createTempFile(name,"shm"), byte_size) : new XMappedFile("/dev/shm/"+name, byte_size);
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    ByteBuffer getBuffer()
    {
        return mMemory.getBuffer();
    }

    void read()
    {
        mMemory.read();
    }

    void write()
    {
        mMemory.write();
    }

}
