/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/

 
package org.jabotics.robot.core;

import org.jabotics.robot.defs.*;
import org.jabotics.robot.util.*;
import org.jabotics.robot.compile.*;
import org.jabotics.robot.adapter.XAdapter;


/**
 * CMotor is the internal controller class for single motors.
 * Higher level motor classes are based on this class.
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class CMotor
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // class variables
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    static private final float ERROR_LIMIT = 10f;
    static private final float MAX_ERROR_TIME = 2f;
    static final float V_MIN = 0.1f;
    static final float A_MIN = 0.1f;
    static final float V_MAX;
    static final float A_MAX = 6000f;
    static final float P_MAX = 100f;
    static private final int NOT_INITIALIZED = -1;
    static private final int UPDATE_PERIOD = 4; // in ms
    static private final int UPDATE_MIN_DELAY = 2; // in ms
    static private final float UPDATE_PERIOD_SEC = 0.001f * UPDATE_PERIOD; // in s
    static private final int CONTROLLER_THREAD_PRIORITY = Thread.MAX_PRIORITY - 2;
    static private final TGate<XMotionDataSet> sGateMDS;
    static private final XMotionDataSet sMDSWrite;
    static private volatile XMotionDataSet sMDSRead;
    static private final CMotor[] sMotors;
    static private final XTimedThread sThread;
    static private final ITimerAdapter sTimer;
    static private double sT = 0.;      // current time
    static private double sTLast = 0.;  // time at last update
    static private float sDT = 0f;      // elapsed time since last update

    // for time dilatation thread:
    static private final int TD_UPDATE_DELAY = 10; // in ms
    static private final float TD_UPDATE_DELAY_F =  (float) TD_UPDATE_DELAY;
    static private final XTimedThread sTDThread;
    static private float sTimeDilatation = 1f;
    static private long sTDLastMillis = 0;


    static
    {
        IAdapterFactory adapterFactory = XAdapter.getFactory();

        // Initialize static variables:
        V_MAX = adapterFactory.getMotorAdapter(EMotorPort.A).getMaxSpeed(); // once initialized at program startup but not updated during runtime

        sGateMDS = new TGate<XMotionDataSet>(XMotionDataSet.PROTOTYPE);
        sMDSWrite = new XMotionDataSet();
        sGateMDS.write(sMDSWrite);
        sMDSRead = sGateMDS.read();
        sMotors = new CMotor[D.MOTOR_PORT_NUMBER];
        for (EMotorPort mp : EMotorPort.values())
        {
            sMotors[mp.ordinal()] = new CMotor(mp,adapterFactory.getMotorAdapter(mp));
        }

        sTimer = adapterFactory.getTimerAdapter();
        sT = 0.;
        sTLast = 0.;

        // Vorbereiten und Starten des Steuerungs-Threads:
        sThread = new XTimedThread("Motor Controller Thread",new ControllerRunnable());
        if (CS.TIMED_THREAD_DILATATION) sThread.enableTimeDilatation(false);
        if (CS.TIMED_THREAD_DEPENDENCIES) sThread.addDependency(D.SESSION_DATA_READER_KEY, 1); // Wait for at least 1 update of simulation dynamics before updating motors.
        sThread.setPriority(CONTROLLER_THREAD_PRIORITY);
        sThread.setDaemon(true);
        sThread.setTimeouts(UPDATE_PERIOD,UPDATE_MIN_DELAY,0);
        sThread.pause(); // initially pausing
        sThread.start();
        if (CS.LOGGING_CORE) XLogBuffer.println("Initialization von class CMotor: thread started");

        if (CS.TIMED_THREAD_DILATATION)
        {
            
            // Vorbereiten und Starten des Zeitdehnungs-Threads:
            sTDThread = new XTimedThread("Time Dilatation Thread",new TimeDilatationRunnable());
            sTDThread.enableTimeDilatation(false);
            sTDThread.setPriority(Thread.NORM_PRIORITY+1);
            //sTDThread.setPriority(Thread.MAX_PRIORITY);
            sTDThread.setDaemon(true);
            sTDThread.setTimeouts(TD_UPDATE_DELAY,TD_UPDATE_DELAY,0);
            sTDThread.pause();
            sTDThread.start();
        }
        else
        {
            sTDThread = null;
        }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // instance variables
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    private final XFloatMotion FLOAT_MOTION = new XFloatMotion();
    private final XHaltedMotion HALTED_MOTION = new XHaltedMotion();
    private final XAcceleratedMotion ACCEL_TO_TARGET = new XAcceleratedMotion();
    private final XAcceleratedMotion ACCEL_TO_HALTED = new XAcceleratedMotion();
    private final XAcceleratedMotion SLOW_DOWN = new XAcceleratedMotion();
    private final XUniformMotion PAUSE_MOTION = new XUniformMotion();
    private final XAcceleratedMotion CORRECTION_1 = new XAcceleratedMotion();
    private final XAcceleratedMotion CORRECTION_2 = new XAcceleratedMotion();
    private final XUniformMotion UNIFORM_MOTION = new XUniformMotion();

    private final XUpperLimitPosition UPPER_LIMIT_POSITION = new XUpperLimitPosition();
    private final XLowerLimitPosition LOWER_LIMIT_POSITION = new XLowerLimitPosition();
    private final XTimeLimit TIME_LIMIT = new XTimeLimit();

    private final int mPort;
    private final IMotorAdapter mMotorAdapter;
    private final float[] mPIDs;

    private final XMotionData mMDWrite;  // motor data that is currently worked on by user (finally to be passed to the motor controller thread)
    private       XMotionData mMDRead;   // last motor data read by motor controller thread (owner of this object)

    private double mTLast;
    private double mTLastCorrection;
    private float  mDT;
    private float  mDTCorrection;
    private float  mPos;
    private float  mPosCorrection;
    private float  mV;
    private float  mVCorrection;
    private float  mPosIs;
    private float  mPowerOut;
    private float  mPowerBase;
    private float  mError1;
    private float  mError2;
    private float  mErrorTime;
    private float  mDTAverage;
    private float  mZeroPos;
    private double mCurrentTargetAbsolute; // used to reinitialize running motion when changing velocity or acceleration
    private XTransaction.MotorCallback mCallback;
    private long   mExternalMotionID;
    private long   mCommandID;      // used to identify new motor transaction (new motion or modified velocity/acceleration)
    private volatile long mLatestMotionID;
    private volatile long mCurrentMotionID;
    private volatile BMotion mMotion;
    private volatile BMotion mMotionCorrection;
    private volatile boolean mIsStalled;
    private volatile float mVTotal;
    private volatile float mPosTotal;

    
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Nested interfaces, classes and enums
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////


    // Appropriate use of class XTransaction will secure proper synchronization for access of class XMotor
    // (compare implementation of CPilot and class XMotor in packages org.jabotics.robot.de/en).
    static public class XTransaction
    {
        enum EState
        {
            RELEASED, WAITING, INTERRUPTED, RUNNING
        }

        private class MotorCallback
        {
            final CMotor mMotor;
            volatile long mTerminatedMotionID = 0;

            MotorCallback(CMotor motor)
            {
                mMotor = motor;
            }

            void motionTerminated(long motionID)
            {
                mTerminatedMotionID = motionID;
            }

            long getMotionID()
            {
                return mMotionID;
            }
        }

        private static XTransaction sFirstWaitingTransaction = null;
        private static XTransaction sLastWaitingTransaction = null;
        private static boolean[] sMotorIsLocked = new boolean[D.MOTOR_PORT_NUMBER];
        private static boolean[] sMotorIsReserved = new boolean[D.MOTOR_PORT_NUMBER];
        static
        {
            for (int i = 0; i < D.MOTOR_PORT_NUMBER; ++i) sMotorIsLocked[i] = false;
        }
        private volatile EState mState = EState.RELEASED;
        private XTransaction mNextWaitingTransaction = null;
        private XTransaction mPrevWaitingTransaction = null;
        private boolean[] mMotorIsUsed = new boolean[D.MOTOR_PORT_NUMBER];
        private long mMotionID = 0;
        private final MotorCallback[] mMotorCBs;

        public XTransaction(CMotor motor)
        {
            mMotorCBs = new MotorCallback[1];
            mMotorCBs[0] = new MotorCallback(motor);
            mMotorIsUsed[motor.mPort] = true;
        }

        public XTransaction(CMotor motor1, CMotor motor2)
        {
            mMotorCBs = new MotorCallback[2];
            mMotorCBs[0] = new MotorCallback(motor1);
            mMotorCBs[1] = new MotorCallback(motor2);
            mMotorIsUsed[motor1.mPort] = true;
            mMotorIsUsed[motor2.mPort] = true;
        }

        public XTransaction(CMotor motor1, CMotor motor2, CMotor motor3)
        {
            mMotorCBs = new MotorCallback[3];
            mMotorCBs[0] = new MotorCallback(motor1);
            mMotorCBs[1] = new MotorCallback(motor2);
            mMotorCBs[2] = new MotorCallback(motor3);
            mMotorIsUsed[motor1.mPort] = true;
            mMotorIsUsed[motor2.mPort] = true;
            mMotorIsUsed[motor3.mPort] = true;
        }

        public EState getState()
        {
            return mState;
        }

        private boolean hasMotorControl(CMotor motor)
        {
            synchronized(mMotorCBs)
            {
                if (EState.INTERRUPTED == mState) return false;
                if (!mMotorIsUsed[motor.mPort]) throw new IllegalStateException("CMotor.XTransaction.hasMotorControl(): Transaction does not have access to motor at port number " + motor.mPort + ".");
                if (EState.RUNNING != mState) throw new IllegalStateException("CMotor.XTransaction.hasMotorControl(): Transaction has not been started yet.");
                return true;
            }
        }

        private boolean tryLock()
        {
            for (int i = 0; i < D.MOTOR_PORT_NUMBER; ++i) if (mMotorIsUsed[i] && sMotorIsLocked[i]) return false;
            for (MotorCallback mcb : mMotorCBs)  sMotorIsLocked[mcb.mMotor.mPort] = true;
            return true;
        }

        private boolean canBeLocked()
        {
            boolean lock = true;
            for (int i = 0; i < D.MOTOR_PORT_NUMBER; ++i)
            {
                if (mMotorIsUsed[i])
                {
                    if (sMotorIsReserved[i])  lock = false;  else  sMotorIsReserved[i] = true;
                }
            }
            return lock;
        }

        private void removeFromWaitingList()
        {
            if (this == sFirstWaitingTransaction)  sFirstWaitingTransaction = mNextWaitingTransaction;
            if (this == sLastWaitingTransaction)  sLastWaitingTransaction  = mPrevWaitingTransaction;
            if (null != mPrevWaitingTransaction)  mPrevWaitingTransaction.mNextWaitingTransaction = mNextWaitingTransaction;
            if (null != mNextWaitingTransaction)  mNextWaitingTransaction.mPrevWaitingTransaction = mPrevWaitingTransaction;
            mNextWaitingTransaction = null;
            mPrevWaitingTransaction = null;
        }

        public void start()
        {
            synchronized(mMotorCBs[0])  // mMotorCBs[0] serves as synchronization object for this motor transaction instance
            {
                synchronized(sGateMDS)  // sGateMDS serves as global synchronization object for motor transactions
                {
                    if (Thread.currentThread().isInterrupted())
                    {
                        mState = EState.INTERRUPTED;
                        return;
                    }
        
                    if (EState.RELEASED != mState) throw new IllegalStateException("CMotor.XTransaction.start(): Transaction already has been started.");
        
                    if (tryLock())
                    {
                        mState = EState.RUNNING;
                        return;
                    }
    
                    // lock failed => add to list of waiting transactions
                    if (null == sFirstWaitingTransaction)
                    {
                        sFirstWaitingTransaction = this;
                    }
                    else
                    {
                        mPrevWaitingTransaction = sLastWaitingTransaction;
                        sLastWaitingTransaction.mNextWaitingTransaction = this;
                    }
                    sLastWaitingTransaction = this;
                }
                try
                {
                    mState = EState.WAITING;
                    mMotorCBs[0].wait();  // mMotorCBs[0] serves as synchronization object for this motor transaction instance
                }
                catch (InterruptedException ie)
                {
                    if (EState.WAITING == mState) // not started yet?
                    {
                        mState = EState.INTERRUPTED;
                    }
                    Thread.currentThread().interrupt();
                }
            }
        }

        public void commit(boolean waitUntilCompleted)
        {
            synchronized(sGateMDS)
            {
                if (EState.INTERRUPTED == mState)
                {
                    this.removeFromWaitingList();
                    mState = EState.RELEASED;
                    return;
                }
    
                if (EState.RUNNING != mState) throw new IllegalStateException("CMotor.XTransaction.commit(): Transaction has not been started yet.");

                for (MotorCallback mcb : mMotorCBs) mcb.mMotor.updateMDForCommit(mcb);
                sGateMDS.write(sMDSWrite);
                for (MotorCallback mcb : mMotorCBs)  sMotorIsLocked[mcb.mMotor.mPort] = false;

                // Wake up waiting transactions with no access conflicts (first in - first out):
                for (int i = 0; i < D.MOTOR_PORT_NUMBER; ++i) sMotorIsReserved[i] = sMotorIsLocked[i];
                XTransaction ta = sFirstWaitingTransaction;
                while(ta != null)
                {
                    XTransaction next = ta.mNextWaitingTransaction;
                    if (ta.canBeLocked())
                    {
                        synchronized(ta.mMotorCBs[0])  // mMotorCBs[0] serves as synchronization object for single motor transaction instances
                        {
                            if (EState.WAITING == ta.mState) // not interrupted yet?
                            {
                                for (MotorCallback mcb : ta.mMotorCBs)  sMotorIsLocked[mcb.mMotor.mPort] = true; // lock motors
                                ta.removeFromWaitingList();
                                ta.mState = EState.RUNNING;
                                ta.mMotorCBs[0].notify();
                            }
                        }
                    }
                    ta = next;
                }

                if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.commit für " + mMotorCBs.length + " Motoren an Port(s) " + mMotorCBs[1].mMotor.mPort + " (...)  ausgeführt.");
        
                if (waitUntilCompleted) waitUntilMotionCompleted();

                mState = EState.RELEASED;
            }
        }
        
        public boolean motionCompleted()
        {
            for (MotorCallback mcb : mMotorCBs) if (mcb.mTerminatedMotionID < mMotionID) return false;
            return true;
        }

        public void waitUntilMotionCompleted()
        {
            while (!motionCompleted())  sTimer.msDelay(1);
        }
    }


    static class TimeDilatationRunnable implements XTimedThread.IRunnable
    {
        private int mCount = 0;

        @Override
        public void init()
        {
            sTDLastMillis = System.currentTimeMillis();
        }

        @Override
        public void runTimed()
        {
            long currentMillis = System.currentTimeMillis();
            float relative_delay = (float)(currentMillis - sTDLastMillis) / TD_UPDATE_DELAY_F - 1f;
            sTimeDilatation = 0.98f * sTimeDilatation + 0.02f *(1f + 4f * relative_delay);
            XTimedThread.setTimeDilatation(sTimeDilatation);
            sTDLastMillis = currentMillis;
            if (CS.LOGGING_TIME_DILATATION)
            {
                if (++mCount > 100)
                {
                    mCount = 0;
                    System.out.println(sTimeDilatation);
                }
                sTDLastMillis = System.currentTimeMillis(); // just in case that logging takes a lot of time
            }
        }

        @Override
        public void terminate()
        {
        }

        @Override
        public void shutdown()
        {
        }
    }


    static class ControllerRunnable implements XTimedThread.IRunnable
    {
        @Override
        public void init()
        {
        }

        @Override
        public void runTimed()
        {
            if (sGateMDS.isModified())
            {
                sMDSRead = sGateMDS.read();
                for (int i = 0; i<D.MOTOR_PORT_NUMBER;  ++i)
                {
                    sMotors[i].mMDRead = sMDSRead.getMDReference(i);
                }
            }

            sTLast = sT;
            sT = 0.001 * (double)sTimer.currentTimeMillis();
            sDT = (float)(sT - sTLast);
            if (sDT < 0) throw new IllegalStateException("CMotor.ControllerRunnable.runTimed(): sDT < 0");
            if (sDT <= 0) return;

            if (!CS.NXT) sMotors[0].mMotorAdapter.freezeAllTachoCounts(); // will result in synchronized position readings for all motors (as far as possible)
            for (CMotor motor : sMotors)
            {
                motor.readCurrentPosition();
            }
            if (!CS.NXT) sMotors[0].mMotorAdapter.releaseAllTachoCounts();
    
            for (CMotor motor : sMotors)
            {
                motor.updateState();
            }

            for (CMotor motor : sMotors)
            {
                motor.setPower();
            }
            if (!CS.NXT) sMotors[0].mMotorAdapter.commitAllPowerData();
        }

        @Override
        public void terminate()
        {
        }

        @Override
        public void shutdown()
        {
            for (CMotor motor : sMotors)
            {
                motor.releaseControl();
            }
        }
    }


    static private class XMotionDataSet implements TICloneable<XMotionDataSet>
    {
        static final XMotionDataSet PROTOTYPE = new XMotionDataSet();

        private final XMotionData[] mMDArray = new XMotionData[D.MOTOR_PORT_NUMBER];
        //volatile VolatileRefArray<XMotionData> mMDArray = new VolatileRefArray<XMotionData>(D.MOTOR_PORT_NUMBER);

        XMotionDataSet()
        {
            for (int i = 0;  i < D.MOTOR_PORT_NUMBER;  ++i)
            {
                mMDArray[i] = new XMotionData();
            }
        }

        XMotionData getMDReference(int port_id)
        {
            return mMDArray[port_id];
        }

        @Override
        public void copyDataFrom(XMotionDataSet template)
        {
            for (int i = 0;  i < D.MOTOR_PORT_NUMBER;  ++i)
            {
                mMDArray[i].copyDataFrom(template.mMDArray[i]);
            }
       }

        // creates clone with new set of motion data
        @Override
        public XMotionDataSet getClone()
        {
            XMotionDataSet mds = new XMotionDataSet();
            mds.copyDataFrom(this);
            return mds;
        }
    }


    static private class XMotionData implements TICloneable<XMotionData>
    {
        static final XMotionData PROTOTYPE = new XMotionData();

        long mCommandID = 0;
        long mMotionID = NOT_INITIALIZED;
        EMotionType mMotionType = null;
        float mVTargetAbs = 360f;       // (in °/s)
        float mATargetAbs = 6000f;      // (in °/s²)
        float mVStartRegular = 0f;      // (in °/s)  start velocity for "regular motion", will be ignored if (mTCorrection == 0)
        float mTStopCorrection = 0f;    // (in s)  time to stop from error velocity (to be corrected)
        float mTarget = 0f;             // target for time/position depending on type of motion
        boolean mTargetIsRelative = true;   // for position target only
        boolean mForward = true;
        XTransaction.MotorCallback mCallback = null;
        long mExternalMotionID = NOT_INITIALIZED;

        @Override
        public void copyDataFrom(XMotionData template)
        {
            if (this != template)
            {
                mCommandID = template.mCommandID;
                mMotionID = template.mMotionID;
                mMotionType = template.mMotionType;
                mVTargetAbs = template.mVTargetAbs;
                mATargetAbs = template.mATargetAbs;
                mVStartRegular = template.mVStartRegular;
                mTStopCorrection = template.mTStopCorrection;
                mTarget = template.mTarget;
                mTargetIsRelative = template.mTargetIsRelative;
                mForward = template.mForward;
                mCallback = template.mCallback;
                mExternalMotionID = template.mExternalMotionID;
            }
        }

        @Override
        public XMotionData getClone()
        {
            XMotionData md = new XMotionData();
            md.copyDataFrom(this);
            return md;
        }
    }

    enum EMotionType
    {
        UNDEFINED, FLOATING, HALTED, UNLIMITED, TIME_LIMITED, POS_LIMITED
    }

    abstract class BMotion
    {
        volatile BMotion mNextMotion = null;
        volatile BMotionStarter mNextMotionStarter = null;
        volatile double mTTarget = 0.;   // target time for this elementary motion
        volatile float mPosTarget = 0f;  // target position for this elementary motion
        volatile float mVTarget = 0f;    // target velocity for this elementary motion
        volatile float mA = 0f;

        void nextMotionStartupCheckAndUpdate()
        {
            if (null != mNextMotionStarter)  mNextMotionStarter.startupCheck();
            mMotion.update();
            if (null != mMotionCorrection) mMotionCorrection.updateCorrection();
            mPosTotal = mPos + mPosCorrection;
            mVTotal = mV + mVCorrection;
            determinePower();
        }

        void determinePower(float P, float I, float D)
        {
            float error = (mPosTotal - mPosIs);
            if (Math.abs(error) > ERROR_LIMIT)
            {
                mErrorTime += sDT;
                if (mErrorTime > MAX_ERROR_TIME)
                {
                    mMotion = FLOAT_MOTION;
                    mErrorTime = 0f;
                    mIsStalled = true;
                    return;
                }
            }
            else
            {
                mErrorTime *= 0.5f;
            }
            ////////////////////////////////////////////////////////////////////////////
            // Ursprünglich ABGESCHAUT IN NXTRegulatedMotor (inzwischen modifiziert)!!!
            ////////////////////////////////////////////////////////////////////////////
            mDTAverage = 0.75f * mDTAverage + 0.25f * sDT;
            float weight = UPDATE_PERIOD_SEC / mDTAverage;
            // use smoothing to reduce the noise in frequent tacho count readings
            // New values
            mError1 = 0.375f * mError1 + 0.625f * error;  // fast smoothing
            mError2 = 0.75f * mError2 + 0.25f * error; // slow smoothing
            // Original values
            //err1 = 0.5f * err1 + 0.5f * error;  // fast smoothing
            //err2 = 0.8f * err2 + 0.2f * error; // slow smoothing
            float power = mPowerBase + (P * mError1 + D * (mError1 - mError2)) * weight;  // normalize response by time period to account for delays in busy system;
            mPowerBase = clipPower(mPowerBase + I * (power - mPowerBase));
            mPowerOut = clipPower(power);
            // Die Instanzvariablen für die Leistung sind vorzeichenbehaftet. Das Vorzeichen gibt die Bewegungsrichtung wieder.
            // Die einfachste Art, das Vorzeichen bei einem Aufruf von MotorPort.controlMotor(int power, int mode) richtig an den Motor durchzuschleusen,
            // besteht darin, unabhängig von der Bewegungsrichtung den Paramter mode mit dem Wert TachoMotorPort.FORWARD zu belegen und die Leistung
            // mit Vorzeichen zu übergeben. Das ist zwar vermutlich nicht ganz die ursprüngliche Intention der Programmierer von TachoMotorPort,
            // aber alles Andere wäre viel zu umständlich. (Vergleiche auch die Implementierungen von MotorPort.controlMotor(int power, int mode)
            // sowie NXTRegulatedMotor.Regulator.calcPower(float error, float P, float I, float D) )
        }

        abstract void update();
        abstract void updateCorrection();
        abstract void determinePower();
        abstract boolean isMoving();
    }

    class XFloatMotion extends BMotion
    {
        void update()
        {
            if (mCallback != null)
            {
                mCallback.motionTerminated(mExternalMotionID);
                mCallback = null;
            }
            float posLast = mPos;
            mPos = 0.8f * mPos + 0.2f * mPosIs;
            if (mDT > 0) mV = (mPos - posLast) / mDT;
        }

        void updateCorrection()
        {
            // nothing to do
        }

        void determinePower()
        {
            mPowerOut = 0f;
        }

        boolean isMoving()
        {
            return Math.abs(mV) > 2;
        }

    }

    class XHaltedMotion extends BMotion
    {
        volatile double mTLastHalted = 0.;   // used to identify restart of halted motion

        void update()
        {
            if (mCallback != null)
            {
                mCallback.motionTerminated(mExternalMotionID);
                mCallback = null;
            }
            if (sTLast != mTLastHalted) mTTarget = sT + 0.1;
            mTLastHalted = sT;
        }

        void updateCorrection()
        {
            // nothing to do
        }

        void determinePower()
        {
            if (sT < mTTarget)
            {
                determinePower(mPIDs[0],mPIDs[1],mPIDs[2]);
            }
            else
            {
                determinePower(mPIDs[3],mPIDs[4],mPIDs[5]);
            }
        }

        boolean isMoving()
        {
            return false;
        }
    }

    class XAcceleratedMotion extends BMotion
    {
        void update()
        {
            float vLast = mV;
            mV += mA * mDT;
            mPos += 0.5f * (vLast + mV) * mDT;
        }

        void updateCorrection()
        {
            float vCorrectionLast = mVCorrection;
            mVCorrection += mA * mDTCorrection;
            mPosCorrection += 0.5f * (vCorrectionLast + mVCorrection) * mDTCorrection;
        }

        void determinePower()
        {
            determinePower(mPIDs[0],mPIDs[1],mPIDs[2]);
        }

        boolean isMoving()
        {
            return true;
        }
    }

    class XUniformMotion extends BMotion
    {
        void update()
        {
            mPos += mV * mDT;
        }

        void updateCorrection()
        {
            // nothing to do
        }

        void determinePower()
        {
            determinePower(mPIDs[0],mPIDs[1],mPIDs[2]);
        }

        boolean isMoving()
        {
            return true;
        }
    }


    abstract class BMotionStarter
    {
        abstract void startupCheck();
    }

    class XUpperLimitPosition extends BMotionStarter
    {
        void startupCheck()
        {
            if ((mV >= 0) && (mMotion.mPosTarget <= mPos))
            {
                mMotion = ACCEL_TO_HALTED;
                float remaining_distance =  mMotion.mPosTarget - mPos;
                if (0f == mV)
                {
                    // Handle this case to avoid division by zero
                    // Actually, should never happen
                    mMotion.mA = -mMDRead.mATargetAbs;
                    mMotion.mTTarget = sT;
                }
                else
                {
                    float remaining_time =  2f * remaining_distance / mV; //  s_remaining / (0.5 * v)
                    if (remaining_time <= 0f)
                    {
                        mMotion.mA = -1000f * mV;
                        mMotion.mTTarget = sT + 0.001;
                    }
                    else
                    {
                        mMotion.mA = -mV / remaining_time;
                        mMotion.mTTarget = sT + remaining_time;
                    }
                }
                // new motion valid starting with next time step => no further update required
            }
        }
    }

    class XLowerLimitPosition extends BMotionStarter
    {
        void startupCheck()
        {
            if ((mV <= 0) && (mMotion.mPosTarget >= mPos))
            {
                mMotion = ACCEL_TO_HALTED;
                float remaining_distance =  mMotion.mPosTarget - mPos;
                if (0f == mV)
                {
                    // Handle this case to avoid division by zero
                    // Actually, should never happen
                    mMotion.mA = mMDRead.mATargetAbs;
                    mMotion.mTTarget = sT;
                }
                else
                {
                    float remaining_time =  2f * remaining_distance / mV; //  s_remaining / (0.5 * v)
                    if (remaining_time <= 0f)
                    {
                        mMotion.mA = -1000f * mV;
                        mMotion.mTTarget = sT + 0.001;
                    }
                    else
                    {
                        mMotion.mA = -mV / remaining_time;
                        mMotion.mTTarget = sT + remaining_time;
                    }
                }
                // new motion valid starting with next time step => no further update required
            }
        }
    }

    class XTimeLimit extends BMotionStarter
    {
        void startupCheck()
        {
            while ((null != mMotionCorrection) && (mMotionCorrection.mTTarget <= sT ))
            {
                mTLastCorrection = mMotionCorrection.mTTarget;
                mDTCorrection = (float)(sT - mTLastCorrection);
                mPosCorrection = mMotionCorrection.mPosTarget;
                mVCorrection = mMotionCorrection.mVTarget;
                mMotionCorrection = mMotionCorrection.mNextMotion;
            }
                
            if (mMotion.mTTarget <= sT )
            {
                mTLast = mMotion.mTTarget;
                mDT = (float)(sT - mTLast);
                mPos = mMotion.mPosTarget;
                mV = mMotion.mVTarget;
                mMotion = mMotion.mNextMotion;
                if (null != mMotion.mNextMotionStarter) mMotion.mNextMotionStarter.startupCheck();
            }
        }
    }


    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Static methods
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////

    // Use this method to get access to an instance of CMotor from higher level motor classes
    static public CMotor getMotor(EMotorPort port)
    {
        CMotor motor = sMotors[port.ordinal()];
        // check if already initialized for use

        synchronized(motor.getClass())
        {
            if (NOT_INITIALIZED == motor.mMDRead.mMotionID)  motor.initialize();
        }
 
        if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.getMotor: Motor für Port " + motor.mPort + " angefordert.");

        return motor;
    }

    static synchronized public void initSession()
    {
        if (CS.TIMED_THREAD_DILATATION) sTimeDilatation = 1f;

        sT = 0.;
        sTLast = 0.;
        sDT = 0f;

        for (CMotor motor : sMotors)
        {
            boolean already_initialized = (motor.mCurrentMotionID != NOT_INITIALIZED);
            motor.initInstanceVariables();
            if (already_initialized) motor.initialize();
        }

        for (int i = 0; i < D.MOTOR_PORT_NUMBER; ++i) XTransaction.sMotorIsLocked[i] = false;
    }

    static synchronized public void startSession()
    {
        if (CS.TIMED_THREAD_DILATATION)
        {
            sTDLastMillis = System.currentTimeMillis();
            sTDThread.resume();
        }
        sThread.resume();
    }

    static public synchronized void pauseSession()
    {
        sThread.pause();
        if (CS.TIMED_THREAD_DILATATION) sTDThread.pause();
    }

    static public synchronized void resumeSession()
    {
        if (CS.TIMED_THREAD_DILATATION) sTDThread.resume();
        sThread.resume();
    }

    static synchronized public void stopSession()
    {
        sThread.pause();
        if (CS.TIMED_THREAD_DILATATION) sTDThread.pause();

        for (CMotor motor : sMotors)
        {
            motor.releaseControl();
        }
    }

    static synchronized public void terminateSession()
    {
        sThread.terminate();
        if (CS.TIMED_THREAD_DILATATION) sTDThread.terminate();
    }


    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Constructors and methods
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////

    private CMotor(EMotorPort port, IMotorAdapter adapter)
    {
        mPort = port.ordinal();
        mMotorAdapter = adapter;
        mPIDs = adapter.getPIDParameters();

        mMDWrite = sMDSWrite.getMDReference(mPort);
         
        ACCEL_TO_HALTED.mNextMotion = HALTED_MOTION;      // ALWAYS!
        ACCEL_TO_HALTED.mNextMotionStarter = TIME_LIMIT;  // ALWAYS!
        SLOW_DOWN.mNextMotionStarter = TIME_LIMIT;        // ALWAYS!
        PAUSE_MOTION.mNextMotion = ACCEL_TO_TARGET;       // ALWAYS!
        PAUSE_MOTION.mNextMotionStarter = TIME_LIMIT;     // ALWAYS!
        CORRECTION_1.mNextMotion = CORRECTION_2;          // ALWAYS!

        initInstanceVariables();
    }

    private void initInstanceVariables()
    {
        mMDRead = XMotionData.PROTOTYPE;

        mTLast = 0.;  // time at last update
        mTLastCorrection = 0.;
        mDT = 0f;      // elapsed time since last update
        mDTCorrection = 0f;
        mPos = 0f;     // calculated position (in °)
        mPosCorrection = 0f;
        mV = 0f;       // calculated velocity (in °/s)
        mVCorrection = 0f;
        mPosIs = 0f;       // measured position (in °)
        mPowerOut = 0f;    // power output
        mPowerBase = 0f;  // integrated power value (PID mechanism)
        mError1 = 0f;     // fast smoothing of tacho count reads in PID mechnaism
        mError2 = 0f;     // slow smoothing of tacho count reads in PID mechnaism
        mErrorTime = 0f;  // motor will turn into float mode if max. error time is exceeded
        mDTAverage = UPDATE_PERIOD_SEC;
        mZeroPos = 0;
        mCurrentTargetAbsolute = 0;
        mCallback = null;
        mExternalMotionID = 0;
        mCommandID = NOT_INITIALIZED;   // used to identify new transactions (new motion/velocity/acceleration)
        mLatestMotionID = NOT_INITIALIZED;  // motion count (latest request by "user")
        mCurrentMotionID = NOT_INITIALIZED; // motion count (currently handled by controller)
        mMotion = null;  // null value means: not in use
        mMotionCorrection = null;  // null value means: no correction
        mIsStalled = false;
        mVTotal = 0f;
        mPosTotal = 0f;
    }

    protected IMotorAdapter getAdapter()
    {
        return mMotorAdapter;
    }

    private void initialize()
    {
        mMotorAdapter.startControl();

        mZeroPos = mMotorAdapter.getTachoCount();

        mMDWrite.copyDataFrom(XMotionData.PROTOTYPE);
        mMDWrite.mMotionType = EMotionType.HALTED;
        mMDWrite.mMotionID = 0; // indicates that motor has been initialized
        mMDRead = sGateMDS.write(sMDSWrite).getMDReference(mPort);
        ++mMDWrite.mCommandID;

        if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.initialize für Motor an Port " + mPort + " ausgeführt.");
    }

    private void readCurrentPosition()
    {
        if (NOT_INITIALIZED != mMDRead.mMotionID) mPosIs = mMotorAdapter.getTachoCount();
    }

    private void updateState()
    {
        if (mMDRead.mMotionID == NOT_INITIALIZED) return; // motor has not been initialized

        // if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.updateState für initialisierten Motor an Port " + mPort + " aufgerufen.");

        mTLast = sTLast;
        mTLastCorrection = sTLast;
        mDT = sDT;
        mDTCorrection = sDT;

        // new motion data available?
        if (mCommandID != mMDRead.mCommandID)  initializeNewMotion();

        if (null != mMotion)  mMotion.nextMotionStartupCheckAndUpdate();

        if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.updateState für Motor an Port " + mPort + ": Zeitparameter:   sTLast: " + sTLast + " ,  sT: " + sT + ",  sDT: " + sDT);
    }

    private void initializeNewMotion()
    {
        if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.initializeNewMotion für Motor an Port " + mPort + " aufgerufen.");

        if (mCurrentMotionID == mMDRead.mMotionID)
        {
            // => Velocity or acceleration change for running motion!
            // No impact if motor is floating / (already) halted:
            if ((mCallback != mMDRead.mCallback) || (mMotion == HALTED_MOTION) || (mMotion == FLOAT_MOTION))
            {
               mCommandID = mMDRead.mCommandID;
               return;
            }
        }
        else
        {
            // new motion
            if (mCallback != null) mCallback.motionTerminated(mExternalMotionID);
            mCallback = mMDRead.mCallback;
            mExternalMotionID = mMDRead.mExternalMotionID;

            mIsStalled = false;
            mPos = mPosTotal;
            mPosCorrection = 0f;
            mV = mVTotal;
            mVCorrection = 0f;
            mMotionCorrection = null;
        }

        switch (mMDRead.mMotionType)
        {
            case HALTED:
            {
                mMotion =  HALTED_MOTION;
                if (CS.ASSERTS_CORE) assert (null == mMotion.mNextMotionStarter) : "CMotor.initializeNewMotion(): HALTED_MOTION.mNextMotionStarter!=null";
                break;
            }
            case UNLIMITED:
            {
                // mCurrentMotionID == mMDRead.mMotionID:
                // Nothing special to do. => Handle like new motion.
                float vTarget = (mMDRead.mForward) ? mMDRead.mVTargetAbs : -mMDRead.mVTargetAbs;
                UNIFORM_MOTION.mNextMotionStarter = null;
                UNIFORM_MOTION.mVTarget = vTarget;
                if (mV == vTarget)
                {
                    mMotion = UNIFORM_MOTION;
                }
                else
                {
                    float dV = vTarget - mV;
                    ACCEL_TO_TARGET.mA = (dV > 0) ?  mMDRead.mATargetAbs : -mMDRead.mATargetAbs;
                    float dT = dV / ACCEL_TO_TARGET.mA;
                    ACCEL_TO_TARGET.mTTarget = sTLast + (double)dT;
                    ACCEL_TO_TARGET.mPosTarget = mPos + 0.5f * (mV + vTarget) * dT;
                    ACCEL_TO_TARGET.mVTarget = vTarget;
                    ACCEL_TO_TARGET.mNextMotionStarter = TIME_LIMIT;
                    ACCEL_TO_TARGET.mNextMotion = UNIFORM_MOTION;
                    mMotion = ACCEL_TO_TARGET;
                }
                if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.initializeNewMotion UNLIMITED für Motor an Port " + mPort + " aufgerufen.");
                break;
            }
            case TIME_LIMITED:
            {
                if (mCurrentMotionID == mMDRead.mMotionID)
                {
                    // Target time is relative and therefore has to be recalculated:
                    mMDRead.mTarget = (float) (mCurrentTargetAbsolute - sTLast);
                }
                else
                {
                    // New motion: memorize absolute target time
                    mCurrentTargetAbsolute = sTLast + mMDRead.mTarget;
                }

                float timeToStopImmediately = (mV < 0) ? -mV / mMDRead.mATargetAbs : mV / mMDRead.mATargetAbs;
                if (mMDRead.mTarget > timeToStopImmediately)
                {
                    ACCEL_TO_TARGET.mNextMotionStarter = TIME_LIMIT;
                    ACCEL_TO_TARGET.mVTarget = (mMDRead.mForward) ? mMDRead.mVTargetAbs : -mMDRead.mVTargetAbs;
                    float dV = ACCEL_TO_TARGET.mVTarget - mV;
                    ACCEL_TO_TARGET.mA = (dV < 0) ?  -mMDRead.mATargetAbs : mMDRead.mATargetAbs;
                    float timeToAccelerate = dV / ACCEL_TO_TARGET.mA;
                    float timeToStop = mMDRead.mVTargetAbs / mMDRead.mATargetAbs;
                    float timeRemaining = mMDRead.mTarget - timeToAccelerate - timeToStop;
                    if (timeRemaining <= 0)
                    {
                        float halfTimeDeficit = 0.5f * timeRemaining;
                        timeToAccelerate += halfTimeDeficit;
                        timeToStop += halfTimeDeficit;
                        ACCEL_TO_TARGET.mTTarget = sTLast + (double)timeToAccelerate;
                        ACCEL_TO_TARGET.mVTarget = mV + ACCEL_TO_TARGET.mA * timeToAccelerate;
                        ACCEL_TO_TARGET.mPosTarget = mPos + 0.5f * (mV + ACCEL_TO_TARGET.mVTarget) * timeToAccelerate;
                        ACCEL_TO_TARGET.mNextMotion = ACCEL_TO_HALTED;
                        ACCEL_TO_HALTED.mPosTarget = ACCEL_TO_TARGET.mPosTarget + 0.5f * ACCEL_TO_TARGET.mVTarget * timeToStop;
                        if (CS.ASSERTS_CORE) assert (Math.abs(sTLast + mMDRead.mTarget - ACCEL_TO_TARGET.mTTarget - timeToStop) < 1e-3) : "CMotor.initializeNewMotion()  for time limited motion: Math.abs(sT + mMDRead.mTarget - ACCEL_TO_TARGET.mTTarget - timeToStop) >= 1e-3";
                    }
                    else
                    {
                        ACCEL_TO_TARGET.mTTarget = sTLast + (double)timeToAccelerate;
                        ACCEL_TO_TARGET.mPosTarget = mPos + 0.5f * (mV + ACCEL_TO_TARGET.mVTarget) * timeToAccelerate;
                        ACCEL_TO_TARGET.mNextMotion = UNIFORM_MOTION;
                        UNIFORM_MOTION.mTTarget = ACCEL_TO_TARGET.mTTarget + (double)timeRemaining;
                        UNIFORM_MOTION.mPosTarget = ACCEL_TO_TARGET.mPosTarget + ACCEL_TO_TARGET.mVTarget * timeRemaining;
                        UNIFORM_MOTION.mVTarget = ACCEL_TO_TARGET.mVTarget;
                        UNIFORM_MOTION.mNextMotionStarter = TIME_LIMIT;
                        UNIFORM_MOTION.mNextMotion = ACCEL_TO_HALTED;
                        ACCEL_TO_HALTED.mPosTarget = UNIFORM_MOTION.mPosTarget + 0.5f * ACCEL_TO_TARGET.mVTarget * timeToStop;
                        if (CS.ASSERTS_CORE) assert (Math.abs(sTLast + mMDRead.mTarget - UNIFORM_MOTION.mTTarget - timeToStop) < 1e-3) : "CMotor.initializeNewMotion()  for time limited motion: Math.abs(sT + mMDRead.mTarget - UNIFORM_MOTION.mTTarget - timeToStop) >= 1e-3";
                    }
                    ACCEL_TO_HALTED.mTTarget = sTLast + (double)mMDRead.mTarget;
                    ACCEL_TO_HALTED.mA = mMDRead.mForward ? -mMDRead.mATargetAbs : mMDRead.mATargetAbs;
                    mMotion = ACCEL_TO_TARGET;
                }
                else
                {
                    // stop immediately
                    ACCEL_TO_HALTED.mTTarget = sTLast + (double)timeToStopImmediately;
                    ACCEL_TO_HALTED.mPosTarget = mPos + 0.5f * mV * timeToStopImmediately;
                    ACCEL_TO_HALTED.mA = (mV < 0) ?  mMDRead.mATargetAbs : -mMDRead.mATargetAbs;
                    mMotion = ACCEL_TO_HALTED;
                }
                if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.initializeNewMotion TIME_LIMITED für Motor an Port " + mPort + " aufgerufen.");
                // Always:
                if (CS.ASSERTS_CORE) assert (UNIFORM_MOTION.mA == 0f) : "CMotor.initializeNewMotion(): UNIFORM_MOTION.mA != 0f";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mNextMotion == HALTED_MOTION) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mNextMotion != HALTED_MOTION";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mNextMotionStarter == TIME_LIMIT) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mNextMotionStarter != TIME_LIMIT";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mVTarget == 0f) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mVTarget != 0f";
                break;
            }
            case POS_LIMITED:
            {
                if (mCurrentMotionID == mMDRead.mMotionID)
                {
                    // No change of velocity / acceleration allowed for "corrected" motions (used by pilot)
                    if (0 < mMDRead.mTStopCorrection) break;
                    
                    // Recalculate relative target position, if necessary:
                    if (mMDRead.mTargetIsRelative)
                    {
                        mMDRead.mTarget = (float) (mCurrentTargetAbsolute - mPosTotal);
                    }
                }
                else
                {
                    // New motion: memorize absolute target position if target is relative:
                    if (mMDRead.mTargetIsRelative) mCurrentTargetAbsolute = mPosTotal + mMDRead.mTarget;
                }
                
                float distance = (mMDRead.mTargetIsRelative) ? mMDRead.mTarget : mMDRead.mTarget - mPos;
                float posStart = mPos;
                float vStart = mV;
                double tStart = sTLast;
                if (0 < mMDRead.mTStopCorrection)
                {
                    mVCorrection = mV - mMDRead.mVStartRegular;
                    mV -= mVCorrection;
                    CORRECTION_1.mVTarget = -D.SQRT05_F * mVCorrection;
                    CORRECTION_1.mA = -mVCorrection / mMDRead.mTStopCorrection;
                    CORRECTION_2.mA = -CORRECTION_1.mA;
                    float tCorrection1 = D.ONE_PLUS_SQRT05_F * mMDRead.mTStopCorrection;
                    float tCorrectionTotal = D.ONE_PLUS_SQRT2_F * mMDRead.mTStopCorrection;
                    CORRECTION_1.mTTarget = sTLast + (double)tCorrection1;
                    CORRECTION_2.mTTarget = sTLast + (double)tCorrectionTotal;
                    CORRECTION_1.mPosTarget = D.ONEHALF_MINUS_ONEHALF_SQRT05_F * mVCorrection * tCorrection1;
                    SLOW_DOWN.mA = (mV < 0) ? mMDRead.mATargetAbs : -mMDRead.mATargetAbs;
                    float tSlowDownToStop = -mV / SLOW_DOWN.mA;
                    float sSlowDown;
                    if (tSlowDownToStop < tCorrectionTotal)
                    {
                        sSlowDown = 0.5f * mV * tSlowDownToStop;
                        SLOW_DOWN.mTTarget = sTLast + tSlowDownToStop;
                        SLOW_DOWN.mVTarget = 0f;
                        SLOW_DOWN.mPosTarget = mPos + sSlowDown;
                        SLOW_DOWN.mNextMotion = PAUSE_MOTION;
                        PAUSE_MOTION.mTTarget = CORRECTION_2.mTTarget;
                        //PAUSE_MOTION.mVTarget = 0f;
                        PAUSE_MOTION.mPosTarget = SLOW_DOWN.mPosTarget;
                        //PAUSE_MOTION.mNextMotion = ACCEL_TO_TARGET;
                    }
                    else
                    {
                        SLOW_DOWN.mTTarget = CORRECTION_2.mTTarget;
                        SLOW_DOWN.mVTarget = mV + SLOW_DOWN.mA * tCorrectionTotal;
                        sSlowDown = 0.5f * (mV + SLOW_DOWN.mVTarget) * tCorrectionTotal;
                        SLOW_DOWN.mPosTarget = mPos + sSlowDown;
                        SLOW_DOWN.mNextMotion = ACCEL_TO_TARGET;
                    }
                    tStart = CORRECTION_2.mTTarget;
                    vStart = SLOW_DOWN.mVTarget;
                    posStart = SLOW_DOWN.mPosTarget;
                    distance -= sSlowDown;
                    mMotion = SLOW_DOWN;
                    mMotionCorrection = CORRECTION_1;
                }
                else
                {
                    mMotion = ACCEL_TO_TARGET;
                }
                float oneOverTwoA = 0.5f / mMDRead.mATargetAbs;
                float direction = (distance < 0) ? -1f : 1f;
                float s = direction * distance;
                float sStop = oneOverTwoA * vStart * vStart;
                float sAccel = oneOverTwoA * mMDRead.mVTargetAbs * mMDRead.mVTargetAbs;
                float tStop = (vStart < 0) ? -vStart / mMDRead.mATargetAbs : vStart / mMDRead.mATargetAbs;
                float tAccel = mMDRead.mVTargetAbs / mMDRead.mATargetAbs;
                float directionCorrection = 1f;
                float sToZero = 0f;  // (virtual) starting position (zero velocity)
                float tToZero = 0f;  // (virtual) starting time (zero velocity)
                float vTargetvStart = direction * mMDRead.mVTargetAbs * vStart;
                if (vTargetvStart > 0)
                {
                    if (sStop > s)
                    {
                        directionCorrection = -1f;
                        sToZero = sStop;
                        tToZero = tStop;
                    } else if (sStop > sAccel)
                    {
                        sToZero = sStop - 2f * sAccel;
                        tToZero = tStop - 2f * tAccel;
                    }
                    else
                    {
                        sToZero = -sStop;
                        tToZero = -tStop;
                    }
                }
                else
                {
                    sToZero = -sStop;
                    tToZero = tStop;
                }
                float sFromZero = (s - sToZero) * directionCorrection;
                if (CS.ASSERTS_CORE) assert (sFromZero >= 0) : "CMotor.initializeNewMotion(): sFromZero < 0";
                float sRemaining = sFromZero - 2f * sAccel;
                ACCEL_TO_TARGET.mVTarget = direction * directionCorrection * mMDRead.mVTargetAbs;
                ACCEL_TO_TARGET.mA = ((ACCEL_TO_TARGET.mVTarget - vStart) < 0) ? -mMDRead.mATargetAbs : mMDRead.mATargetAbs;
                if (sRemaining > 0)
                {
                    ACCEL_TO_TARGET.mNextMotion = UNIFORM_MOTION;
                    ACCEL_TO_TARGET.mNextMotionStarter = TIME_LIMIT;
                    ACCEL_TO_TARGET.mTTarget = tStart + (double)(tToZero + tAccel);
                    ACCEL_TO_TARGET.mPosTarget = posStart + direction * (sToZero + directionCorrection * sAccel);
                    UNIFORM_MOTION.mNextMotion = ACCEL_TO_HALTED;
                    UNIFORM_MOTION.mNextMotionStarter = TIME_LIMIT;
                    UNIFORM_MOTION.mTTarget = ACCEL_TO_TARGET.mTTarget + (double)(sRemaining / mMDRead.mVTargetAbs);
                    UNIFORM_MOTION.mPosTarget = ACCEL_TO_TARGET.mPosTarget + direction * directionCorrection * sRemaining;
                    UNIFORM_MOTION.mVTarget = ACCEL_TO_TARGET.mVTarget;
                    ACCEL_TO_HALTED.mTTarget = UNIFORM_MOTION.mTTarget + (double)tAccel;
                    //if (CS.ASSERTS_CORE) assert (Math.abs(UNIFORM_MOTION.mPosTarget + direction * directionCorrection * sAccel - ACCEL_TO_HALTED.mPosTarget) < 1) : "CMotor.initializeNewMotion()  for position limited motion: Math.abs(UNIFORM_MOTION.mPosTarget + direction * directionCorrection * sAccel - ACCEL_TO_HALTED.mPosTarget) < 1";
                }
                else
                {
                    ACCEL_TO_TARGET.mNextMotion = ACCEL_TO_HALTED;
                    ACCEL_TO_TARGET.mNextMotionStarter = (ACCEL_TO_TARGET.mVTarget < 0) ? LOWER_LIMIT_POSITION : UPPER_LIMIT_POSITION;
                    ACCEL_TO_TARGET.mPosTarget = posStart + direction * (sToZero + directionCorrection * (sAccel+0.5f*sRemaining));
                }
                ACCEL_TO_HALTED.mPosTarget = posStart + distance;
                ACCEL_TO_HALTED.mA = (ACCEL_TO_TARGET.mVTarget < 0) ? mMDRead.mATargetAbs : -mMDRead.mATargetAbs;
                // Always:
                if (CS.ASSERTS_CORE) assert (UNIFORM_MOTION.mA == 0f) : "CMotor.initializeNewMotion(): UNIFORM_MOTION.mA != 0f";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mNextMotion == HALTED_MOTION) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mNextMotion != HALTED_MOTION";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mNextMotionStarter == TIME_LIMIT) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mNextMotionStarter != TIME_LIMIT";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mVTarget == 0f) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mVTarget != 0f";
                break;
            }
            default:
            {
                mMotion =  FLOAT_MOTION;
                if (CS.ASSERTS_CORE) assert (null == mMotion.mNextMotionStarter) : "CMotor.initializeNewMotion(): FLOAT_MOTION.mNextMotionStarter != null";
            }
        }

        mCommandID = mMDRead.mCommandID;
        mCurrentMotionID = mMDRead.mMotionID;
    }

    /*private void determinePower()
    {
        //if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.determinePower für Motor an Port " + mPort + " aufgerufen.");
        if (null != mMotion) mMotion.determinePower();
    }*/

    private void setPower()
    {
        if (null == mMotion) return;
        mMotorAdapter.setPower(mPowerOut);
        //if (CS.LOGGING_CORE)  XLogBuffer.println("CMotor.setPower:   Port: " + mPort + "   Leistung: " + mPowerOut + " ,   PosSoll: " + mPos + " ,   PosIst: " + mPosIs + " ,   Tacho (Motor): " + mMotorAdapter.getTachoCount());
    }

    static private float clipPower(float power)
    {
        return (P_MAX < power) ? P_MAX : ((-P_MAX > power) ? -P_MAX : power);
    }

    private void releaseControl()
    {
        if (null != mMotion)
        {
            mMotion = null;
            mMotionCorrection = null;
            mMotorAdapter.setToFloatMode();
        }
    }

    private void updateMDForCommit(XTransaction.MotorCallback callback)
    {
        if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.updateMDForCommit für Motor an Port " + mPort + " aufgerufen.");
        ++mMDWrite.mCommandID;
        mMDWrite.mCallback = callback;
        mMDWrite.mExternalMotionID = callback.getMotionID();
        mLatestMotionID = mMDWrite.mMotionID;
    }

    // liefert den aktuellen Soll-Winkelstand der Motorsteuerung, nicht den wirklich Ist-Stand von mMotorAdapter.getTachoCount()
    public float getPosition()
    {
        return (mPosTotal - mZeroPos);
    }

    // liefert den Soll-Winkelstand der Motorsteuerung ohne Berücksichtigung des aktuellen Nullpunkts des Motors
    public float getInternalPosition()
    {
        return mPosTotal;
    }

    public void resetPosition()
    {
        mZeroPos = mPosTotal;
    }

    public float getSpeed()
    {
        return mVTotal;
    }

    public float getTargetSpeed()
    {
        return mMDWrite.mVTargetAbs;
    }

    // It's OK to pass negative value for speed!
    public void setTargetSpeed(XTransaction ta, float speed)
    {
        if (!ta.hasMotorControl(this)) return;
        mMDWrite.mVTargetAbs = Math.min(V_MAX, Math.max(V_MIN, Math.abs(speed)));
    }

    public float getTargetAcceleration()
    {
        return mMDWrite.mATargetAbs;
    }

    public void setTargetAcceleration(XTransaction ta, float acceleration)
    {
        if (!ta.hasMotorControl(this)) return;
        mMDWrite.mATargetAbs = Math.min(A_MAX, Math.max(A_MIN, Math.abs(acceleration)));
    }

    public float getMaxSpeed()
    {
        return V_MAX;
    }

    public void setTargetSpeedToMax(XTransaction ta)
    {
        setTargetSpeed(ta,V_MAX);
    }

    public float getMaxAcceleration()
    {
        return A_MAX;
    }

    public void setTargetAccelerationToMax(XTransaction ta)
    {
        setTargetAcceleration(ta,A_MAX);
    }

    public void setTachoCountCorrectionFactor(float correctionFactor)
    {
        mMotorAdapter.setTachoCountCorrectionFactor(correctionFactor);
    }

    public void startRotationUnlimited(XTransaction ta, boolean forward)
    {
        if (!ta.hasMotorControl(this)) return;
        ++ta.mMotionID; // multiple increments by different motors do not matter
        ++mMDWrite.mMotionID;
        mMDWrite.mMotionType = EMotionType.UNLIMITED;
        mMDWrite.mForward = forward;
    }

    public void startRotationTimeLimited(XTransaction ta, float time_ms, boolean forward)
    {
        if (!ta.hasMotorControl(this)) return;
        ++ta.mMotionID; // multiple increments by different motors do not matter
        ++mMDWrite.mMotionID;
        mMDWrite.mMotionType = EMotionType.TIME_LIMITED;
        mMDWrite.mTarget = 0.001f * time_ms;
        mMDWrite.mForward = forward;
    }

    public void startRotationPosLimited(XTransaction ta, float position, boolean isRelative)
    {
        if (!ta.hasMotorControl(this)) return;
        ++ta.mMotionID; // multiple increments by different motors do not matter
        ++mMDWrite.mMotionID;
        mMDWrite.mMotionType = EMotionType.POS_LIMITED;
        mMDWrite.mTStopCorrection = 0f;
        mMDWrite.mTarget = (isRelative) ? position : position + mZeroPos; // absolute position: external value -> internal value
        mMDWrite.mTargetIsRelative = isRelative;
    }

    public void startRotationPosLimited(XTransaction ta, float distance, float v_start_regular, float t_stop_correction)
    {
        if (!ta.hasMotorControl(this)) return;
        ++ta.mMotionID; // multiple increments by different motors do not matter
        ++mMDWrite.mMotionID;
        mMDWrite.mMotionType = EMotionType.POS_LIMITED;
        mMDWrite.mVStartRegular = v_start_regular;
        mMDWrite.mTStopCorrection = t_stop_correction;
        mMDWrite.mTarget = distance;
        mMDWrite.mTargetIsRelative = true;
    }

    public void stopRotation(XTransaction ta)
    {
        startRotationTimeLimited(ta, 0f, true);
    }

    public void setToFloatMode(XTransaction ta)
    {
        if (!ta.hasMotorControl(this)) return;
        ++ta.mMotionID; // multiple increments by different motors do not matter
        ++mMDWrite.mMotionID;
        mMDWrite.mMotionType = EMotionType.FLOATING;
    }

    public boolean isMoving()
    {
        if (mLatestMotionID > mCurrentMotionID) return true;
        BMotion motion = mMotion;
        return ((null != motion) && (motion.isMoving()));
    }

    public boolean isStalled()
    {
        return mIsStalled;
    }
}
