/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.de;

import org.jabotics.robot.defs.*;
import org.jabotics.robot.core.CMotor;

/**
 * Objekte vom Typ XMotor repräsentieren einen Motor des NXT-Roboters.
 * Die Klasse XMotor stellt dem Nutzer die Schnittstelle
 * {@link IMotor} zur Ansteuerung zur Verfügung.
 * Der Konstruktor der Klasse ist "nicht öffentlich". Man kann deshalb
 * als Anwender selbst keine Objekte dieses Typs erzeugen.
 * Man verschafft sich Zugang zu einem neuen Motor-Objekt, indem man die Methode
 * {@link IObjektfabrik#erzeugeMotor()} aufruft.
 * Man erhält Zugriff auf die einzige Objektfabrik-Instanz, indem man
 * die statische Methode {@link XObjektfabrik#holeObjektfabrik()} aufruft.
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */
class XMotor implements IMotor
{
    // Instanzvariablen
    private final CMotor mMotor;
    private final CMotor.XTransaction mMotorTransaction;

    XMotor(EMotoranschluss anschluss)
    {
        mMotor = CMotor.getMotor(EMotorPort.values()[anschluss.ordinal()]);
        mMotorTransaction = new CMotor.XTransaction(mMotor);
    }

    @Override
    public void setzeKorrekturfaktorWinkelstand(float korrekturfaktor)
    {
        mMotor.setTachoCountCorrectionFactor(korrekturfaktor);
    }

    @Override
    public float holeWinkelstand()
    {
        return mMotor.getPosition();
    }

    @Override
    public void setzeWinkelstandZurueck()
    {
        mMotorTransaction.start();
        mMotor.resetPosition();
        mMotorTransaction.commit(false);
    }
    
    @Override
    public float holeGeschwindigkeit()
    {
        return mMotor.getSpeed();
    }
    
    @Override
    public float holeSollgeschwindigkeit()
    {
        return mMotor.getTargetSpeed();
    }
    
    @Override
    public void setzeSollgeschwindigkeit(float winkelgeschwindigkeit)
    {
        mMotorTransaction.start();
        mMotor.setTargetSpeed(mMotorTransaction,winkelgeschwindigkeit);
        mMotorTransaction.commit(false);
    }
    
    @Override
    public float holeSollbeschleunigung()
    {
        return mMotor.getTargetAcceleration();
    }
    
    @Override
    public void setzeSollbeschleunigung(float winkelbeschleunigung)
    {
        mMotorTransaction.start();
        mMotor.setTargetAcceleration(mMotorTransaction,winkelbeschleunigung);
        mMotorTransaction.commit(false);
    }

    @Override
    public float holeMaximalgeschwindigkeit()
    {
        return mMotor.getMaxSpeed();
    }
    
    @Override
    public void setzeSollgeschwindigkeitMaximal()
    {
        mMotorTransaction.start();
        mMotor.setTargetSpeedToMax(mMotorTransaction);
        mMotorTransaction.commit(false);
    }
    
    @Override
    public float holeMaximalbeschleunigung()
    {
        return mMotor.getMaxAcceleration();
    }
    
    @Override
    public void setzeSollbeschleunigungMaximal()
    {
        mMotorTransaction.start();
        mMotor.setTargetAccelerationToMax(mMotorTransaction);
        mMotorTransaction.commit(false);
    }
    
    @Override
    public void drehe(ERichtung richtung)
    {
        mMotorTransaction.start();
        mMotor.startRotationUnlimited(mMotorTransaction,ERichtung.VORWAERTS == richtung);
        mMotorTransaction.commit(false);
    }
    
    @Override
    public void dreheZeit(float dauer, ERichtung richtung, EWartemodus wartemodus)
    {
        mMotorTransaction.start();
        mMotor.startRotationTimeLimited(mMotorTransaction, dauer, (ERichtung.VORWAERTS == richtung));
        mMotorTransaction.commit(EWartemodus.ABWARTEN == wartemodus);
    }

    @Override
    public void dreheUmWinkel(float drehwinkel, EWartemodus wartemodus)
    {
        mMotorTransaction.start();
        mMotor.startRotationPosLimited(mMotorTransaction, drehwinkel, true);
        mMotorTransaction.commit(EWartemodus.ABWARTEN == wartemodus);
    }

    @Override
    public void dreheZuWinkelstand(float winkelstand, EWartemodus wartemodus)
    {
        mMotorTransaction.start();
        mMotor.startRotationPosLimited(mMotorTransaction, winkelstand, false);
        mMotorTransaction.commit(EWartemodus.ABWARTEN == wartemodus);
    }

    @Override
    public void warteEndeAb()
    {
        mMotorTransaction.waitUntilMotionCompleted();
    }
    
    @Override
    public void halteAn(EWartemodus wartemodus)
    {
        if (mMotor.isMoving())
        {
            mMotorTransaction.start();
            mMotor.stopRotation(mMotorTransaction);
            mMotorTransaction.commit(EWartemodus.ABWARTEN == wartemodus);
        }
    }
    
    @Override
    public void gebeFrei()
    {
        mMotorTransaction.start();
        mMotor.setToFloatMode(mMotorTransaction);
        mMotorTransaction.commit(false);
    }
    
    @Override
    public boolean istInBewegung()
    {
        return mMotor.isMoving();
    }
    
    @Override
    public boolean istAusgesetzt()
    {
        return mMotor.isStalled();
    }
}
