/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.util;

import org.jabotics.robot.compile.*;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * Die Klasse XTimedThread bietet einen Mechanismus zum regelmäßigen zeitgesteuerten Aufruf einer
 * runTimed()-Methode der an den Konstruktor von XTimedThread übergebenen XTimedThread.Runnable-Instanz.
 * Mit einem setPeriods(long period,long min_period)-Aufruf kann das Zeitintervall vorgegeben werden,
 * in dem die runTimed()-Methode aufgerufen wird, wobei min_period das minimale Zeitintervall ist,
 * das nach Durchlaufen der runTimed-Methode vor einem erneuten Aufruf abgewartet wird.
 * Vor dem ersten Aufruf von runTimed() wird das Runnable-Objekt durch Aufruf der init()-Methode initialisiert.
 * Dieser Aufruf erfolgt aus dem Timer-Thread heraus, so dass der Thread mit Thread.currentThread()
 * z.B. zur späteren Verwendung in terminate() (Aufruf aus einem anderen Thread!) ermittelt werden kann.
 * Die runTimed()-Methode kann auch gezielt durch einen Aufruf der Methode trigger() angestoßen werden.
 * Durch Aufruf der Methode terminate() wird der Thread geordnet gestoppt. Dabei wird die Runnable-Instanz
 * durch Aufruf der terminate()-Methode davon in Kenntnis gesetzt, dass der Thread gestoppt werden soll.
 * Die Runnable-Instanz kann so bei Bedarf eine beschleunigte Beendigung eines runTimed()-Durchlaufs veranlassen.
 * Es wird von XTimedThread zusätzlich ein Interrupt auf den Thread ausgelöst für den Fall, dass sich
 * der Thread in einem Wartezustand befindet. Nach Beendigung der Timer-Schleife wird die shutdown()-Methode
 * der Runnable-Instanz aufgerufen, so dass hier notwendige Aufräumarbeiten erledigt werden können.
 * Spätestens bei Programmende werden diese Terminierungsschritte druchlaufen.
 * 
 * Implemention has to take into account limitations of class java.lang.Thread for NXT.
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class XTimedThread
{
    public interface IRunnable
    {
        void init();
        void runTimed();
        void terminate();
        void shutdown();
    }

    static public abstract class BMonitor
    {
        private final String mKey;

        protected BMonitor(String key)
        {
            mKey = key;
            registerThreadMonitor(this);
        }

        final public String getKey() {return mKey;}
        final protected void unregister() {unregisterThreadMonitor(mKey);};

        abstract public long getCalls();
        
        @Override
        public boolean equals(Object object)
        {
            if (this == object) return true;
            if ((null == object) || !(object instanceof BMonitor)) return false;
            return mKey.equals(((BMonitor)object).mKey);
        }

        @SuppressWarnings("deprecation")
        @Override
        protected void finalize()
        {
            unregisterThreadMonitor(mKey);
        }
    }

    static private class Monitor extends BMonitor
    {
        private volatile long mCalls = 0;
        
        Monitor(String key)
        {
            super(key);
        }
 
        @Override
        public long getCalls()
        {
            return mCalls;
        }
    }

    private class Run implements java.lang.Runnable
    {
        public void run()
        {
            XTimedThread.this.run();
        }
    }

    private class ShutdownHook extends Thread
    {
        @Override
        public void run()
        {
            terminate();
        }
    }

    private static class Dependency
    {
        private final String      mKey;
        private volatile BMonitor mMonitor;
        private volatile int      mMinCalls;
        private long              mTargetCalls;

        private Dependency(String monitor_key, int min_calls)
        {
            mKey = monitor_key;
            mMonitor = null;
            mMinCalls = (min_calls < 1) ? 1 : min_calls;
            mTargetCalls = mMinCalls;
        }

        private Dependency(BMonitor monitor, int min_calls)
        {
            mKey = monitor.mKey;
            mMonitor = monitor;
            mMinCalls = (min_calls < 1) ? 1 : min_calls;
            mTargetCalls = mMinCalls;
        }
    }

    // constants / class variables:
    private static final long  MILLION = 1000000;
    //private static final float MILLION_F = MILLION;
    private static final long  sStartNanos = System.nanoTime();
    // class variables for load balancing mechanism:
    private static final ArrayList<BMonitor> sRegisteredThreadMonitors;
    private static final ArrayList<Dependency> sPendingDependencies;
    private static volatile float sTimeDilatation = 1f;
    static
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            sRegisteredThreadMonitors = new ArrayList<BMonitor>();
            sPendingDependencies = new ArrayList<Dependency>();
        }
        else
        {
            sRegisteredThreadMonitors = null;
            sPendingDependencies = null;
        }
    }
    
    // Die Instanzvariablen:
    private final Thread mThread = new Thread(new Run());
    private final Object mSynchObject = new Object();
    private final Object mStateMonitor = new Object();
    private final Object mRunMonitor = new Object();
    private final ShutdownHook mShutdownHook = new ShutdownHook();
    private LinkedList<Dependency> mDependencies = (CS.TIMED_THREAD_DEPENDENCIES ? new LinkedList<Dependency>() : null);
    private volatile Monitor   mMonitor = null;
    private volatile IRunnable mRunnable = null;
    private volatile boolean   mTimeDilatationEnabled = CS.TIMED_THREAD_DILATATION;
    private volatile boolean   mStarted = false;
    private volatile boolean   mPausing = false;
    private volatile boolean   mTerminate = false;
    private volatile boolean   mTrigger = false;
    private volatile long      mTargetTime = 0;
    private volatile int       mPeriodMillis = 10;
    private volatile int       mMinDelayMillis = 5;
    private volatile int       mInitialDelayMillis = 0;
    private volatile long      mPeriodL;
    private volatile long      mMinDelayL;
    private volatile long      mInitialDelayL;
    private volatile float     mPeriodF;
    private volatile float     mMinDelayF;
    private volatile float     mInitialDelayF;

    public XTimedThread(IRunnable runnable)
    {
        if (null == runnable) throw new IllegalArgumentException("XTimedThread: Runnable is null");
        mRunnable = runnable;
        initTimeouts();
    }

    public XTimedThread(String key, IRunnable runnable)
    {
        if (null == key) throw new IllegalArgumentException("XTimedThread: Key is null");
        if (null == runnable) throw new IllegalArgumentException("XTimedThread: Runnable is null");
        mRunnable = runnable;
        mThread.setName(key);
        if (CS.TIMED_THREAD_DEPENDENCIES) mMonitor = new Monitor(key); // Constructor will register monitor instance
        initTimeouts();
    }

    private void initTimeouts()
    {
        if (CS.TIMED_THREAD_NANOSECONDS)
        {
            mPeriodL = MILLION * (long)mPeriodMillis;
            mMinDelayL = MILLION * (long)mMinDelayMillis;
            mInitialDelayL = MILLION * (long)mInitialDelayMillis;
        }
        else
        {
            mPeriodL = (long)mPeriodMillis;
            mMinDelayL = (long)mMinDelayMillis;
            mInitialDelayL = (long)mInitialDelayMillis;
        }

        mPeriodF = (float)mPeriodL;
        mMinDelayF = (float)mMinDelayL;
        mInitialDelayF = (float)mInitialDelayL;
    }

    public String getKey()
    {
        return (mMonitor == null) ? null : ((BMonitor)mMonitor).mKey;
    }

    public String getName()
    {
        return mThread.getName();
    }

    public void setName(String name)
    {
        mThread.setName(name);
    }

    public int getPriority()
    {
        return mThread.getPriority();
    }

    public void setPriority(int priority)
    {
        mThread.setPriority(priority);
    }

    public void setDaemon(boolean on)
    {
        mThread.setDaemon(on);
    }

    public boolean timeDilatationEnabled()
    {
        return mTimeDilatationEnabled;
    }

    public void enableTimeDilatation(boolean enabled)
    {
        if (CS.TIMED_THREAD_DILATATION) mTimeDilatationEnabled = enabled;
    }

    public void start()
    {
        if (!mStarted)
        {
            synchronized(mStateMonitor)
            {
                if (!mStarted)
                {
                    Runtime.getRuntime().addShutdownHook(mShutdownHook);
                    mThread.start();
                    mStarted = true;
                }
            }
        }
    }

    public void pause()
    {
        synchronized(mStateMonitor)
        {
            synchronized(mRunMonitor) // synchronization to make sure that timed thread is really pausing after this call
            {
                mPausing = true;
            }
        }
    }

    public void resume()
    {
        synchronized(mStateMonitor)
        {
            mPausing = false;
        }
    }

    public void terminate()
    {
        if (mStarted && !mTerminate)
        {
            synchronized(mStateMonitor)
            {
                if (!mTerminate)
                {
                    mRunnable.terminate();
                    mTerminate = true;
                    if (mThread.isAlive())
                    {
                        mThread.interrupt();
                        try
                        {
                            mThread.join();
                        }
                        catch (InterruptedException e)
                        {
                            Thread.currentThread().interrupt();
                        }
                    }
                }
            }
        }
    }

    public void trigger()
    {
        synchronized(mSynchObject)
        {
            mTrigger = true;
            mSynchObject.notify();
        }
    }

    public int getPeriod()
    {
        return mPeriodMillis;
    }

    public int getMinDelay()
    {
        return mMinDelayMillis;
    }

    public int getInitialDelay()
    {
        return mInitialDelayMillis;
    }

    public void setTimeouts(int period, int min_delay, int initial_delay)
    {
        synchronized(mSynchObject)
        {
            mPeriodMillis = (period > 0) ? period : 1;
            mMinDelayMillis = (min_delay < 0) ? 0 : min_delay;
            mInitialDelayMillis = (initial_delay < 0) ? 0 : initial_delay;

            initTimeouts();
        }
    }

    public BMonitor getMonitor()
    {
        return mMonitor;
    }

    public void addDependency(String monitor_key, int min_calls)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            synchronized(mDependencies)
            {
                if (null == mDependencies) return;

                Iterator<Dependency> it_dep = mDependencies.iterator();
                while (it_dep.hasNext())
                {
                    Dependency dep = it_dep.next();
                    if (dep.mKey.equals(monitor_key))
                    {
                        dep.mMinCalls = (min_calls < 1) ? 1 : min_calls;
                        return;
                    }
                }
                mDependencies.add(getDependency(monitor_key, min_calls));
            }
        }
    }

    public void addDependency(XTimedThread thread, int min_calls)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            BMonitor monitor = thread.getMonitor();
            if (null == monitor) throw new IllegalArgumentException("XTimedThread.addDependency: Invalid thread parameter. Thread needs to own a thread monitor. Specify a key on thread construction.");
            addDependency(monitor, min_calls);
        }
    }

    public void addDependency(BMonitor monitor, int min_calls)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            if (null == mDependencies) return;

            synchronized(mDependencies)
            {
                Iterator<Dependency> it_dep = mDependencies.iterator();
                while (it_dep.hasNext())
                {
                    Dependency dep = it_dep.next();
                    if (monitor.equals(dep.mMonitor))
                    {
                        dep.mMinCalls = (min_calls < 1) ? 1 : min_calls;
                        return;
                    }
                }
                mDependencies.add(new Dependency(monitor, min_calls));
            }
        }
    }

    public void clearDependencies()
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            synchronized(mDependencies)
            {
                if (null == mDependencies) return;

                mDependencies.clear();
                mDependencies = null;
            }
        }
    }

    private void run()
    {
        try
        {
            mRunnable.init();

            if (CS.TIMED_THREAD_NANOSECONDS)
            {
                mTargetTime = nanoTime();
            }
            else
            {
                mTargetTime = System.currentTimeMillis();
            }

            if (mInitialDelayL > 0)
            {
                if (CS.TIMED_THREAD_DILATATION)
                {
                    mTargetTime += (mTimeDilatationEnabled) ? (long)(sTimeDilatation * mInitialDelayF) : mInitialDelayL;
                }
                else
                {
                    mTargetTime += mInitialDelayL;
                }
                waitUntilTargetTime();
            }
            callRunTimed(); // initial call
            
            while(!mTerminate && !mThread.isInterrupted())
            {
                synchronized(mSynchObject)
                {
                    if (!mTrigger)  waitUntilTargetTime();
                }
                callRunTimed();
            }
        }
        catch(InterruptedException e)
        {
            if (CS.LOGGING_UTIL)
            {
                if (XLogBuffer.isEnabled())
                {
                    synchronized(XLogBuffer.getMonitor())
                    {
                        XLogBuffer.print("InterruptedException in XTimedThread: ");
                        XLogBuffer.println(this.mThread.getName());
                        XLogBuffer.print("Message: ");
                        XLogBuffer.println(e.getMessage());
                    }
                }
            }
        }

        mRunnable.shutdown();

        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            clearDependencies();
        }
    }

    private void waitUntilTargetTime() throws InterruptedException
    {
        if (CS.TIMED_THREAD_NANOSECONDS)
        {
            long currentNanos = nanoTime();
            if (mTargetTime > currentNanos)
            {
                long timeout = mTargetTime - currentNanos;
                int nanos = (int)(timeout % MILLION);
                long millis = (timeout - nanos) / MILLION;
                synchronized(mSynchObject)
                {
                    mSynchObject.wait(millis,nanos);
                }
            }
        }
        else
        {
            long currentMillis = System.currentTimeMillis();
            if (mTargetTime > currentMillis)
            {
                synchronized(mSynchObject)
                {
                    mSynchObject.wait(mTargetTime - currentMillis);
                }
            }
        }
    }

    private void callRunTimed() throws InterruptedException
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            Iterator<Dependency> it_dep = mDependencies.iterator();
            while (it_dep.hasNext())
            {
                Dependency dep = it_dep.next();
                // If thread monitor has been registered (that is: not pending), check dependency
                if (null != dep.mMonitor)
                {
                    while (dep.mTargetCalls > dep.mMonitor.getCalls())
                    {
                        synchronized(mSynchObject)
                        {
                            mSynchObject.wait(1);
                        }
                    }
                    dep.mTargetCalls = dep.mMonitor.getCalls() + dep.mMinCalls;
                }
            }
        }

        if (CS.LOGGING_UTIL)
        {
            if (XLogBuffer.isEnabled())
            {
                synchronized(XLogBuffer.getMonitor())
                {
                    XLogBuffer.print(this.mThread.getName());
                    XLogBuffer.print(": runTimed() at ");
                    XLogBuffer.println(String.valueOf(mTargetTime));
                }
            }
        }
        mTrigger = false;
        synchronized(mRunMonitor)
        {
            if (!mPausing)
            {
                mRunnable.runTimed();
                if (CS.TIMED_THREAD_DEPENDENCIES)  if (null != mMonitor) ++mMonitor.mCalls;
            }
        }
        long minTargetTime;
        if (CS.TIMED_THREAD_NANOSECONDS)
        {
            minTargetTime = nanoTime();
        }
        else
        {
            minTargetTime = System.currentTimeMillis();
        }
        if (CS.TIMED_THREAD_DILATATION)
        {
            float dilatation = sTimeDilatation;
            if (mTimeDilatationEnabled && (dilatation > 1f))
            {
                mTargetTime += (long)(dilatation * mPeriodF);
                minTargetTime += (long)(dilatation * mMinDelayF);
            }
            else
            {
                mTargetTime += mPeriodL;
                minTargetTime += mMinDelayL;
            }
        }
        else
        {
            mTargetTime += mPeriodL;
            minTargetTime += mMinDelayL;
        }
        if (mTargetTime < minTargetTime) mTargetTime = minTargetTime;
    }

    static public float getTimeDilatation()
    {
        return sTimeDilatation;
    }

    static public void setTimeDilatation(float dilatation)
    {
        sTimeDilatation = (dilatation < 1f) ? 1f : dilatation;
    }

    static private void registerThreadMonitor(BMonitor monitor)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            synchronized (sRegisteredThreadMonitors)
            {
                String key = monitor.mKey;
                Iterator<BMonitor> it_monitor = sRegisteredThreadMonitors.iterator();
                while (it_monitor.hasNext())
                {
                    if (key.equals(it_monitor.next().mKey)) throw new IllegalArgumentException("Keys of XTimedThread instances must be unique or null. Key: " + key);
                }
                sRegisteredThreadMonitors.add(monitor);
    
                Iterator<Dependency> it_dep = sPendingDependencies.iterator();
                while (it_dep.hasNext())
                {
                    Dependency dep = it_dep.next();
                    if (key.equals(dep.mKey))
                    {
                        dep.mMonitor = monitor;
                        it_dep.remove();
                    }
                }
            }
        }
    }

    static private void unregisterThreadMonitor(String key)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            synchronized (sRegisteredThreadMonitors)
            {
                Iterator<BMonitor> it = sRegisteredThreadMonitors.iterator();
                while (it.hasNext())
                {
                    BMonitor monitor = it.next();
                    if (monitor.mKey.equals(key))
                    {
                        it.remove();
                        return;
                    }
                }
            }
        }
    }

    static private Dependency getDependency(String key, int min_calls)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            synchronized (sRegisteredThreadMonitors)
            {
                Iterator<BMonitor> it = sRegisteredThreadMonitors.iterator();
                while (it.hasNext())
                {
                    BMonitor monitor = it.next();
                    if (key.equals(monitor.mKey)) return new Dependency(monitor,min_calls);
                }
    
                Dependency dep = new Dependency(key,min_calls);
                sPendingDependencies.add(dep);
                return dep;
            }
        }
        else
        {
            return null;
        }
    }

    static private long nanoTime()
    {
        return System.nanoTime() - sStartNanos;
    }
}
