/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;


import java.io.IOException;
import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicLong;

import org.jabotics.pc.defs.SIMDefs;
import org.jabotics.robot.compile.CS;
import org.jabotics.robot.defs.*;
import org.jabotics.robot.util.XLogBuffer;
import org.jabotics.robot.util.XTimedThread;

/**
 * SIMAdapter implements all adapter interfaces for running robot programs within the robot simulation tool.
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class SIMAdapter implements IAdapterFactory, ITimerAdapter, IDisplayAdapter, ISoundAdapter, IConsoleAdapter, IRobotAdapter, ISessionAdapter
{
    static
    {
        // Initialization of constants defined in class Env depend on context.
        // => Set EnvInit.CONTEXT before accessing constants in class Env.
        EnvInit.CONTEXT = EContext.ROBOT;
        // Initialize class Env if not yet done:
        Env.init();
    }

    static private final long MILLION = 1000000;
    static private final int UPDATE_PERIOD = 2; // in ms
    static private final int UPDATE_MIN_DELAY = 1; // in ms

    static private ESimulationMode sSimulationMode = ESimulationMode.TEST; // default value
    static private SIMAdapter sSIMAdapter = null;

    // Set simulation mode prior to calling getAdapterFactory() 
    static public synchronized ESimulationMode setSimulationMode(ESimulationMode mode)
    {
        if ((null == sSIMAdapter) && Env.ROBOT_IS_MASTER)
        {
            if (ESimulationMode.NORMAL == mode) mode = ESimulationMode.TEST; // Mode "NORMAL" not supported in this case
            sSimulationMode = mode;
        }
        return sSimulationMode;
    }

    static public synchronized SIMAdapter getSIMAdapter()
    {
        if (null == sSIMAdapter) sSIMAdapter = new SIMAdapter();
        return sSIMAdapter;
    }


    private class SessionDataReader extends XTimedThread.BMonitor implements XTimedThread.IRunnable
    {
        SessionDataReader()
        {
            super(D.SESSION_DATA_READER_KEY);
        }

        @Override
        public long getCalls()
        {
            return mSessionData.mDynamicsUpdateCount;
        }

        private void proceedToState(ESessionState targetState)
        {
            synchronized(mSessionStateMonitor)
            {
                if (ESessionState.TERMINATING == mCurrentState) return;

                boolean terminating = (ESessionState.TERMINATING == targetState);
    
                if (terminating) targetState = ESessionState.COMPLETED;
                
                while (targetState != mCurrentState)
                {
                    switch (mCurrentState)
                    {
                    case NEW:
                    case COMPLETED:
                        initRobotSession();
                        mCurrentState = ESessionState.INITIALIZED;
                        break;
                    case INITIALIZED:
                        startRobotSession();
                        mCurrentState = ESessionState.RUNNING;
                        break;
                    case PAUSING:
                        resumeRobotSession();
                        mCurrentState = ESessionState.RUNNING;
                        break;
                    case RUNNING:
                        if (ESessionState.PAUSING == targetState)
                        {
                            pauseRobotSession();
                            mCurrentState = ESessionState.PAUSING;
                        }
                        else
                        {
                            stopRobotSession();
                            mCurrentState = ESessionState.STOPPED;
                        }
                        break;
                    case STOPPED:
                        mCurrentState = ESessionState.COMPLETED;
                        break;
                    default:
                        throw new IllegalStateException("SIMAdapter.SessionDataReader.proceedToState: Not possible to procceed from " + mCurrentState.toString() + " to " + targetState.toString() + ".");
                     }
                }

                if (terminating)
                {
                    terminateRobotSession();
                    mCurrentState = ESessionState.TERMINATING;

                    // Inform the simulation that the robot execution has ended,
                    // because the simulation will wait for all robot sessions to terminate:
                    mSimulation.sessionRobotExecutionEnded(mSessionID);

                    // If robot and simulation are running in the same process,
                    // leave it to the simulation to terminate the process!
                    if (Env.ROBOT_RUNS_IN_SEPARATE_PROCESS)
                    {
                        shutdown();

                        System.exit(0);
                    }
                }

                mSessionStateMonitor.notifyAll();
            }
        }

        @Override
        public synchronized void runTimed()
        {
            mSessionData = new SessionData(new XSessionData(mSimulation.sessionGetData(mSessionID)));
            
            if (mCurrentState != mSessionData.mSessionState) proceedToState(mSessionData.mSessionState);
        }

       @Override
       public void init() {}

        @Override
        public void terminate() {}

        @Override
        public void shutdown() {}
        
    }

    private class SessionData
    {
        ESessionState mSessionState;
        long mTRobotStart;
        long mDynamicsUpdateCount;
        
        // Default constructor corresponds to initial session state
           SessionData()
        {
            mSessionState = ESessionState.NEW;
            mTRobotStart = 0;
            mDynamicsUpdateCount = 0;
        }

           SessionData(XSessionData session_data)
        {
            mSessionState = session_data.getSesstionState();
            mTRobotStart = System.nanoTime() - session_data.getTRobot();
            mDynamicsUpdateCount = session_data.getDynamicsUpdateCount();
        }
    }


    private final Thread mRobotThread = Thread.currentThread();
    private final Object mSessionStateMonitor = new Object();
    private final Object mRobotThreadMonitor = new Object();
    private boolean mSuspendRobotThreadIsEnabled;
    private volatile boolean mRobotThreadIsWaiting = false;
    private int mSessionID;
    private ISimulation mSimulation = null;
    private Process mSimulationProcess = null;
    private SessionDataReader mSessionDataReader;
    private XTimedThread mSessionDataReaderThread = null;
    private final AtomicLong mTRobot = new AtomicLong(0);
    private volatile boolean mPausing = true;
    private volatile SessionData mSessionData = new SessionData();
    private ESessionState mCurrentState = ESessionState.NEW;

    
    private final SensorAdapter mInvalidSensorAdapter = new SensorAdapter(new XSensorInfo(-1,ESensorType.INVALID_SENSOR.ordinal(),1,false));
    private SensorAdapter[] mSensors = new SensorAdapter[D.TOTAL_SENSOR_NUMBER];
    private float[] mSensorValues = new float[D.TOTAL_SENSOR_NUMBER];

    private MotorAdapter[] mMotors = new MotorAdapter[D.MOTOR_PORT_NUMBER];
    private float[] mTachoCounts = new float[D.MOTOR_PORT_NUMBER];
    private float[] mPowerValues = new float[D.MOTOR_PORT_NUMBER];

    private SIMAdapter() // bare construction needs to be followed by initialize() directly after initializing XAdapterFactory.sFactory
    {
    }

    public void initialize()
    {

        if (CS.LOG_BUFFER)
        {
            if (Env.ROBOT_RUNS_IN_SEPARATE_PROCESS) XLogBuffer.enableLogging(true);
        }

        if (Env.ROBOT_IS_MASTER)
        {
            mSessionID = 0;

            if (CS.LOGGING_SIM) XLogBuffer.println("SIMAdapter.initialize(): Robot is master => Session id is 0.");
        }
        else
        {
            String sessionID = System.getProperty(SIMDefs.KEY_SIM_SESSION_ID);
            if (null == sessionID)
            {
                throw new InternalError("Internal error in org.jabotics.robot.sim.SIMAdapter: Session identifier undefined.");
            }
            try
            {
                mSessionID = Integer.parseInt(sessionID);
            }
            catch (NumberFormatException e)
            {
                throw new InternalError("Internal error in org.jabotics.robot.sim.SIMAdapter: Session identifier not properly defined.");
            }

            if (CS.LOGGING_SIM) XLogBuffer.println("SIMAdapter.initialize(): Robot is not master. Session id is " + mSessionID + ".");
        }

        if (Env.ROBOT_IS_MASTER)
        {
            // SIMAdapter is responsible for setting simulation mode
            // prior to starting the simulation.

            if (Env.ROBOT_RUNS_IN_SEPARATE_PROCESS)
            {
                if (CS.LOGGING_SIM) XLogBuffer.println("SIMAdapter.initialize(): Robot process will start simulation process now in selected simulation mode.");

                // Start simulation process (thereby passing simulation mode and uuid):
                String java = System.getProperty("java.home") + (Env.WINDOWS ? "\\bin\\java.exe" : "/bin/java");
                ArrayList<String> commandList = new ArrayList<String>();
                ProcessBuilder pb = new ProcessBuilder(commandList);
                pb.redirectOutput(ProcessBuilder.Redirect.INHERIT);
                commandList.add(java);
                commandList.add("-D" + SIMDefs.KEY_SIM_UUID + "=" + Env.SIM_UUID);
                commandList.add("-D" + SIMDefs.KEY_SIM_MODE + "=" + sSimulationMode.ordinal());
                commandList.add("-D" + SIMDefs.KEY_RMI_PORT + "=" + Env.RMI_PORT);
                commandList.add("-cp");
                commandList.add(System.getProperty("java.class.path"));
                commandList.add("org.jabotics.robot.sim.XSimulation");
                try
                {
                    mSimulationProcess = pb.start();
                }
                catch (IOException e)
                {
                    throw new InternalError("SIMAdapter.initianlize: Failed to start simulation process. Simulation tool needs to be in the classpath of the current process.");
                }
            }
            else
            {
                if (CS.LOGGING_SIM) XLogBuffer.println("SIMAdapter.initialize(): Set simulation mode (simulation running in robot process).");

                XSimulation.setMode(sSimulationMode);
            }
        }

        if (Env.ROBOT_RUNS_IN_SEPARATE_PROCESS)
        {
            if (CS.LOGGING_SIM) XLogBuffer.println("SIMAdapter.initialize(): Simulation server will be instantiated.");

            mSimulation = new XSimulationProxy();
        }
        else
        {
            if (CS.LOGGING_SIM) XLogBuffer.println("SIMAdapter.initialize(): Simulation will be instantiated.");

            mSimulation = XSimulation.getSimulation();
        }

        for (int i = 0; i < D.MOTOR_PORT_NUMBER; ++i) mMotors[i] = new MotorAdapter(i);

        mSuspendRobotThreadIsEnabled = mSimulation.sessionSuspendRobotThreadIsEnabled(mSessionID);

        mSessionDataReader = new SessionDataReader();
        mSessionDataReaderThread = new XTimedThread("Session data reader thread", mSessionDataReader);
        mSessionDataReaderThread.setPriority(Thread.MAX_PRIORITY-1);
        mSessionDataReaderThread.setTimeouts(UPDATE_PERIOD,UPDATE_MIN_DELAY,0);
        mSessionDataReaderThread.setDaemon(true);
        // Start session data reader
        boolean interactive = ((ESimulationMode.NORMAL == sSimulationMode) || (ESimulationMode.TEST == sSimulationMode));
        if (interactive)
        {
            synchronized (mRobotThreadMonitor)
            {
                mRobotThreadIsWaiting = true;
                mSessionDataReaderThread.start();
                try
                {
                    // Let this thread wait until session is being started via the UI and the session state
                    // changes to RUNNING (see resumeRobotThread() which is called in startRobotSession()).
                    mRobotThreadMonitor.wait();
                }
                catch (InterruptedException e) {}
            }
        }
        else
        {
            // Thread must not be halted because it will control the session state programmatically. 
            mSessionDataReaderThread.start();
        }

        if (Env.ROBOT_IS_MASTER && Env.ROBOT_RUNS_IN_SEPARATE_PROCESS)
        {
            // Make sure that simulation is informed about end of robot execution.
            // 1) !Env.ROBOT_IS_MASTER || !Env.ROBOT_RUNS_IN_SEPARATE_PROCESS:
            //    Nothing to do because simulation will directly monitor related
            //    processes or threads.
            // 2) Otherwise: Robot program will be executed only once in this process.
            //    Create joining thread to inform simulation about end of robot execution.
            (new Thread() {
                public void run()
                {
                    try
                    {
                        mRobotThread.join();
                    }
                    catch (InterruptedException e) {}
    
                    mSimulation.sessionRobotExecutionEnded(mSessionID);
                }
            }).start();
        }
    }

    public void shutdown()
    {
        if (ESessionState.TERMINATING != mCurrentState)
        {
            synchronized(mSessionStateMonitor)
            {
                mSessionDataReader.proceedToState(ESessionState.TERMINATING);
            }
        }

        if (mSimulation instanceof XSimulationProxy)
        {
            // separate processes

            ((XSimulationProxy) mSimulation).shutdownConnection();

            if (Env.ROBOT_IS_MASTER)
            {
                try
                {
                    // Wait for forked simulation process to terminate
                    mSimulationProcess.waitFor();
                }
                catch (InterruptedException e) {}
            }
        }
    }

    @Override
    protected void finalize() throws Throwable
    {
        if (null != mSimulation) mSimulation = null;
        super.finalize();
    }

    @SuppressWarnings("deprecation")
    private void suspendRobotThread()
    {
        if (mSuspendRobotThreadIsEnabled) mRobotThread.suspend();
    }

    @SuppressWarnings("deprecation")
    private void resumeRobotThread()
    {
        synchronized(mRobotThreadMonitor)
        {
            if (mRobotThreadIsWaiting)
            {
                mRobotThreadMonitor.notifyAll();
                mRobotThreadIsWaiting = false;
            }
            else
            {
                if (mSuspendRobotThreadIsEnabled) mRobotThread.resume();
            }
        }
    }

    private class SensorAdapter implements ISensorAdapter
    {
        XSensorInfo mInfo;

        private SensorAdapter(XSensorInfo info)
        {
            mInfo = info;
        }

        @Override
        public int getPortID()
        {
            return mInfo.mPortID;
        }

        @Override
        public ESensorType getType()
        {
            return ESensorType.values()[mInfo.mSensorTypeID];
        }

        @Override
        public int getUpdatePeriodProlongationFactor()
        {
            return mInfo.mUpdatePeriodProlongationFactor;
        }

        @Override
        public float getValue()
        {
            return mSensorValues[mInfo.mPortID];
        }

        @Override
        public boolean hasLight()
        {
            return mInfo.mHasLight;
        }

        @Override
        public void lightOn()
        {
            if (mInfo.mHasLight) mSimulation.sensorSwitchLightOn(mSessionID, mInfo.mPortID);
        }

        @Override
        public void lightOff()
        {
            if (mInfo.mHasLight) mSimulation.sensorSwitchLightOff(mSessionID, mInfo.mPortID);
        }

        @Override
        public void prepareSensorUpdates()
        {
            mSensorValues = mSimulation.sensorGetValues(mSessionID);
        }
    }


    private class MotorAdapter implements IMotorAdapter
    {
        private final int mPortID;
        
        private MotorAdapter(int port_id)
        {
            mPortID = port_id;
        }

        @Override
        public float getMaxSpeed()
        {
            return mSimulation.motorGetMaxSpeed(mSessionID, mPortID);
        }

        @Override
        public void startControl()
        {
            mSimulation.motorStartControl(mSessionID, mPortID);
        }

        @Override
        public void setToFloatMode()
        {
            mSimulation.motorSetToFloatMode(mSessionID, mPortID);
        }

        @Override
        public void setPower(float power)
        {
            mPowerValues[mPortID] = power;
        }

        @Override
        public float getTachoCount()
        {
            return mTachoCounts[mPortID];
        }

        @Override
        public float[] getPIDParameters()
        {
            return mSimulation.motorGetPIDParameters(mSessionID, mPortID);
        }

        @Override
        public void setTachoCountCorrectionFactor(float correctionFactor)
        {
        }

        @Override
        public void freezeAllTachoCounts()
        {
            mTachoCounts = mSimulation.motorGetTachoCounts(mSessionID);
        }

        @Override
        public void releaseAllTachoCounts()
        {
        }

        @Override
        public void commitAllPowerData()
        {
            mSimulation.motorSetAllPowerData(mSessionID, mPowerValues);
        }
    }
    

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // methods of IAdapterFactory
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public IMotorAdapter getMotorAdapter(EMotorPort port)
    {
        return mMotors[port.ordinal()];
    }

    @Override
    public ISensorAdapter getSensorAdapter(ESensorPort port, ESensorType type)
    {
        int port_id = port.ordinal();
        int type_id = type.ordinal();
        SensorAdapter sensor = mSensors[port_id]; 
        if (null != sensor)
        {
            return (sensor.mInfo.mSensorTypeID == type_id) ? sensor : null;
        }
        else
        {
            XSensorInfo info = null;
            info = mSimulation.sensorRegisterSensor(mSessionID, port_id, type_id);
            if (null == info)
            {
                throw new IllegalStateException("SIMAdapter: Different sensors assigned to the same sensor port. Check robot implementation.");
            }
            else
            {
                sensor = new SensorAdapter(info);
                mSensors[port_id] = sensor;
                return sensor;
            }
        }
    }

    @Override
    public ISensorAdapter getButtonAdapter(EButtonType type)
    {
        int type_id = type.ordinal();
        int port_id = D.SENSOR_PORT_NUMBER + type_id;
        SensorAdapter sensor = mSensors[port_id]; 
        if (null != sensor)
        {
            return sensor;
        }
        else
        {
            XSensorInfo info = null;
            info = mSimulation.sensorRegisterButton(mSessionID, type_id);
            if (null == info)
            {
                throw new IllegalStateException("Internal error in SIMAdapter: Failed to get button adapter.");
            }
            else
            {
                sensor = new SensorAdapter(info);
                mSensors[port_id] = sensor;
                return sensor;
            }
        }
    }

    @Override
    public ISensorAdapter getInvalidSensorAdapter()
    {
        return mInvalidSensorAdapter;
    }

    @Override
    public ISoundAdapter getSoundAdapter()
    {
        return this;
    }

    @Override
    public IDisplayAdapter getDisplayAdapter()
    {
        return this;
    }

    @Override
    public ITimerAdapter getTimerAdapter()
    {
        return this;
    }

    @Override
    public IConsoleAdapter getConsoleAdapter()
    {
        return this;
    }

    @Override
    public IRobotAdapter getRobotAdapter()
    {
        return this;
    }

    @Override
    public ISessionAdapter getSessionAdapter()
    {
        return this;
    }


    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // methods of ITimerAdapter
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public long currentTimeMillis()
    {
        return nanoTimeRobot() / MILLION;
    }

    @Override
    public void msDelay(long period)
    {
        nsDelayRobot(period * MILLION);
    }

    private long nanoTimeRobot()
    {
        SessionData currentData = mSessionData;
        if ((null == currentData) || mPausing) return mTRobot.get();
        long extrapolatedNanos = System.nanoTime() - currentData.mTRobotStart;
        long localNanos = mTRobot.get();
        while (extrapolatedNanos > localNanos)
        {
            mTRobot.compareAndSet(localNanos, extrapolatedNanos);
            localNanos = mTRobot.get();
        }
        return localNanos;
    }

    private void nsDelayRobot(long period)
    {
        long remaining = period;
        long start = nanoTimeRobot();
        while (remaining > 0)
        {
            nsDelaySystem(remaining);
            if (Thread.currentThread().isInterrupted()) return;
            remaining = period - (nanoTimeRobot() - start);
        }
    }

    void nsDelaySystem(long period)
    {
        if (period > 0)
        {
            int nanos = (int)(period % MILLION);
            long millis = (period - nanos) / MILLION;
            try
            {
                Thread.sleep(millis,nanos);
            }
            catch (InterruptedException e)
            {
                Thread.currentThread().interrupt();
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // methods of IDisplayAdapter
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void refresh()
    {
        mSimulation.displayRefresh(mSessionID);
    }

    @Override
    public void clear()
    {
        mSimulation.displayClear(mSessionID);
    }

    @Override
    public void clear(int y_position)
    {
        mSimulation.displayClear(mSessionID, y_position);
    }

    @Override
    public void clear(int x_position, int y_position, int count)
    {
        mSimulation.displayClear(mSessionID, x_position, y_position, count);
    }

    @Override
    public void draw(char ch, int x_position, int y_position)
    {
        mSimulation.displayDraw(mSessionID, ch, x_position, y_position);
    }

    @Override
    public void draw(int n, int x_position, int y_position)
    {
        mSimulation.displayDraw(mSessionID, n, x_position, y_position);
    }

    @Override
    public void draw(int n, int x_position, int y_position, int digits)
    {
        mSimulation.displayDraw(mSessionID, n, x_position, y_position, digits);
    }

    @Override
    public void draw(float fl, int x_position, int y_position)
    {
        mSimulation.displayDraw(mSessionID, fl, x_position, y_position);
    }

    @Override
    public void draw(String text, int x_position, int y_position)
    {
        mSimulation.displayDraw(mSessionID, text, x_position, y_position);
    }

    @Override
    public void draw(String text, int x_position, int y_position, boolean inverted)
    {
        mSimulation.displayDraw(mSessionID, text, x_position, y_position, inverted);
    }

    
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // methods of ISoundAdapter
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public int getVolume()
    {
        return mSimulation.soundGetVolume(mSessionID);
    }

    @Override
    public void setVolume(int volume)
    {
        mSimulation.soundSetVolume(mSessionID, volume);
    }

    @Override
    public void beep()
    {
        msDelay(mSimulation.soundBeep(mSessionID));
    }

    @Override
    public void twoBeeps()
    {
    	msDelay(mSimulation.soundTwoBeeps(mSessionID));
    }

    @Override
    public void beepSequence()
    {
    	msDelay(mSimulation.soundBeepSequence(mSessionID));
    }

    @Override
    public void beepSequenceUp()
    {
    	msDelay(mSimulation.soundBeepSequenceUp(mSessionID));
    }

    @Override
    public void buzz()
    {
    	msDelay(mSimulation.soundBuzz(mSessionID));
    }

    @Override
    public void tone(int frequency, int period)
    {
        mSimulation.soundTone(mSessionID, frequency, period);
        msDelay(period);
    }

    @Override
    public void tone(int frequency, int period, int volume)
    {
        mSimulation.soundTone(mSessionID, frequency, period, volume);
        msDelay(period);
    }

    @Override
    public void note(ESound instrument, int frequency, int period)
    {
        mSimulation.soundNote(mSessionID, instrument.ordinal(), frequency, period);
        msDelay(period);
    }

    @Override
    public void pause(int period)
    {
        msDelay(period);
    }

    
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // methods of IConsoleAdapter
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void open()
    {
        mSimulation.consoleOpen(mSessionID);
    }

    @Override
    public void close()
    {
        mSimulation.consoleClose(mSessionID);
    }

    @Override
    public void print(String text)
    {
        mSimulation.consolePrint(mSessionID, text);
    }

    @Override
    public void println(String text)
    {
        mSimulation.consolePrintln(mSessionID, text);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // methods of IRobotAdapter
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public float getDefaultTrackWidth()
    {
        return mSimulation.robotGetDefaultTrackWidth(mSessionID);
    }

    @Override
    public float getDefaultWheelDiameter()
    {
        return mSimulation.robotGetDefaultWheelDiameter(mSessionID);
    }

    @Override
    public void setTrackWidth(float trackWidth)
    {
        mSimulation.robotSetTrackWidth(mSessionID, trackWidth);
    }

    @Override
    public void setWheelDiameter(float wheel_diameter)
    {
        mSimulation.robotSetWheelDiameter(mSessionID, wheel_diameter);
    }

    @Override
    public void setMotors(EMotorPort leftMotorPort, EMotorPort rightMotorPort)
    {
        mSimulation.robotSetMotors(mSessionID, leftMotorPort.ordinal(), rightMotorPort.ordinal());
    }


    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // methods of ISessionAdapter for programmatic session control of a simulation run in test or demo mode
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void resetPreferences()
    {
        mSimulation.sessionResetPreferences(mSessionID);
    }

    @Override
    public void importPreferences(Class<?> cl)
    {
        String name = cl.getName();
        mSimulation.sessionImportPreferencesByClass(mSessionID, name);
    }

    @Override
    public void importPreferences(String pathname)
    {
        mSimulation.sessionImportPreferences(mSessionID, pathname);
    }

    @Override
    public void pressButton(EButtonType type, int period)
    {
        mSimulation.sessionPressButton(mSessionID, type.ordinal(), period);
    }

    @Override
    public void clap(int volume)
    {
        mSimulation.sessionClap(volume);
    }

    @Override
    public void setNoiseLevel(int level)
    {
        mSimulation.sessionSetNoiseLevel(level);
    }

    @Override
    public void startSession()
    {
        mSimulation.sessionInit(mSessionID);
        waitUntilStateProgagatedBack(ESessionState.INITIALIZED);
        mSimulation.sessionStart(mSessionID);
        waitUntilStateProgagatedBack(ESessionState.RUNNING);
    }

    @Override
    public void stopSession()
    {
        mSimulation.sessionStop(mSessionID);
        waitUntilStateProgagatedBack(ESessionState.STOPPED);
        mSimulation.sessionComplete(mSessionID);
        waitUntilStateProgagatedBack(ESessionState.COMPLETED);
    }

    private void waitUntilStateProgagatedBack(ESessionState state)
    {
        synchronized(mSessionStateMonitor)
        {
            try
            {
                while (state != mCurrentState) mSessionStateMonitor.wait();
            }
            catch (InterruptedException e) {}
        }
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // methods for robot side session handling
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    public void initRobotSession()
    {
        mTRobot.set(0);
        
        mSensorValues = new float[D.TOTAL_SENSOR_NUMBER];

        mTachoCounts = new float[D.MOTOR_PORT_NUMBER];
        mPowerValues = new float[D.MOTOR_PORT_NUMBER];

        org.jabotics.robot.core.CSessionManager.initSession();
    }

    public void startRobotSession()
    {
        mPausing = false;

        org.jabotics.robot.core.CSessionManager.startSession();

        resumeRobotThread();
    }

    public void pauseRobotSession()
    {
        suspendRobotThread();

        org.jabotics.robot.core.CSessionManager.pauseSession();

        mPausing = true;
    }

    public void resumeRobotSession()
    {
        mPausing = false;

        org.jabotics.robot.core.CSessionManager.resumeSession();

        resumeRobotThread();
    }

    public void stopRobotSession()
    {
        suspendRobotThread();

        org.jabotics.robot.core.CSessionManager.stopSession();

        mPausing = true;
    }

    public void terminateRobotSession()
    {
        resumeRobotThread(); // main thread should better not be suspended when terminating

        org.jabotics.robot.core.CSessionManager.terminateSession();
    }
}
