/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import java.rmi.Naming;
import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.concurrent.atomic.AtomicInteger;

import org.jabotics.robot.compile.CS;
import org.jabotics.robot.util.XLogBuffer;

/**
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class XSimulationProxy implements ISimulation
{
    private ISimulationServer mSimulationServer = null;
    private AtomicInteger mUserCount = new AtomicInteger(0);
    private volatile boolean mTerminating = false;
    private final Object mUsageMonitor = new Object();
    private final Object mThreadGetter = new Object();


    XSimulationProxy()
    {
        for (int i = 0; i < 100; ++i)
        {
            SIMAdapter.getSIMAdapter().nsDelaySystem(100000000);
            try
            {
                mSimulationServer = (ISimulationServer)Naming.lookup(Env.RMI_SERVER);
            }
            catch (Exception e) {}

            if (null != mSimulationServer) break;
        }
        if (null == mSimulationServer) throw new IllegalStateException("XSimulationProxy: Failed to connect to simulation server. Please restart simulation tool.");

        String simulationUUIDVerification = null;
        try
        {
            simulationUUIDVerification = mSimulationServer.simulationGetUUID();
        }
        catch (RemoteException e) {simulationAborted();}

        if (!Env.SIM_UUID.equals(simulationUUIDVerification))
        {
            mSimulationServer = null;
            throw new IllegalStateException("XSimulationProxy: Failed to connect to simulation server. Please restart simulation tool.");
        }

        if (CS.LOGGING_SIM) XLogBuffer.println("Succeeded to load simulation stub.");
    }

    public void shutdownConnection()
    {
        mTerminating = true;
        
        synchronized(mUsageMonitor)
        {
            // wait until no usages left
            try
            {
                if (mUserCount.get() > 0) mUsageMonitor.wait();
            }
            catch (InterruptedException e) {Thread.currentThread().interrupt();}
        }
        
        try
        {
            UnicastRemoteObject.unexportObject(mSimulationServer, true);
        }
        catch (NoSuchObjectException e) {}

    }

    private void simulationAborted()
    {
        if (!mTerminating) throw new IllegalStateException("XSimulationProxy: Remote exception in robot process. Simulation aborted?");
    }

    /*private void enterMethod()
    {
        if (mTerminating)
        {
            // Current thread is to be halted by throwing an exception:
            throw new IllegalStateException("XSimulationProxy: Terminating.");
        }

        mUserCount.incrementAndGet();

        // Check again if terminating
        if (mTerminating)
        {
            // Undo increment operation (now synchronized):
            synchronized(mUsageMonitor)
            {
                if (0 == mUserCount.decrementAndGet())
                {
                    mUsageMonitor.notifyAll();
                }
            }

            // Current thread is to be halted by throwing an exception:
            throw new IllegalStateException("XSimulationProxy: Terminating.");
        }
    }

    private void leaveMethod()
    {
        mUserCount.decrementAndGet();

        if (mTerminating)
        {
            synchronized(mUsageMonitor)
            {
                if (0 == mUserCount.get())
                {
                    mUsageMonitor.notifyAll();
                }
            }

            // Current thread is to be halted by throwing an exception:
            throw new IllegalStateException("XSimulationProxy: Terminating.");
        }
    }*/

    private void enterMethod()
    {
        if (mTerminating)
        {
            // Current thread will be halted before method call to simulation server.
            try
            {
                mThreadGetter.wait();
            }
            catch (InterruptedException e) {Thread.currentThread().interrupt();}
        }

        mUserCount.incrementAndGet();

        // Check again if terminating
        if (mTerminating)
        {
            // Undo increment operation (now synchronized):
            synchronized(mUsageMonitor)
            {
                if (0 == mUserCount.decrementAndGet())
                {
                    mUsageMonitor.notifyAll();
                }
            }

            // Current thread will be halted before method call to simulation server.
            try
            {
                mThreadGetter.wait();
            }
            catch (InterruptedException e) {Thread.currentThread().interrupt();}
        }
    }

    private void leaveMethod()
    {
        mUserCount.decrementAndGet();

        if (mTerminating)
        {
            synchronized(mUsageMonitor)
            {
                if (0 == mUserCount.get())
                {
                    mUsageMonitor.notifyAll();
                }
            }
        }
    }

    /*********************************************************************************************
     * Implementation of ISimulation
     *********************************************************************************************/

    @Override
    public String simulationGetUUID()
    {
        enterMethod();
        String uuid = null;
        try
        {
            uuid = mSimulationServer.simulationGetUUID();
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return uuid;
    }

    @Override
    public boolean sessionSuspendRobotThreadIsEnabled(int session_id)
    {
        enterMethod();
        boolean enabled = false;
        try
        {
            enabled = mSimulationServer.sessionSuspendRobotThreadIsEnabled(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return enabled;
    }

    @Override
    public long[] sessionGetData(int session_id)
    {
        enterMethod();
        long[] data = null;
        try
        {
            data = mSimulationServer.sessionGetData(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return data;
    }

    @Override
    public void sessionResetPreferences(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionResetPreferences(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sessionImportPreferences(int session_id, String pathname)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionImportPreferences(session_id, pathname);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sessionImportPreferencesByClass(int session_id, String classname)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionImportPreferencesByClass(session_id, classname);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sessionInit(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionInit(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sessionStart(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionStart(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sessionStop(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionStop(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sessionComplete(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionComplete(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sessionTerminate(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionTerminate(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public synchronized void sessionRobotExecutionEnded(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionRobotExecutionEnded(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sessionPressButton(int session_id, int button_id, int period)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionPressButton(session_id, button_id, period);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sessionClap(int volume)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionClap(volume);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sessionSetNoiseLevel(int volume)
    {
        enterMethod();
        try
        {
            mSimulationServer.sessionSetNoiseLevel(volume);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public XSensorInfo sensorRegisterSensor(int session_id, int port_id, int type_id)
    {
        enterMethod();
        XSensorInfo info = null;
        try
        {
            info = mSimulationServer.sensorRegisterSensor(session_id, port_id, type_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return info;
    }

    @Override
    public XSensorInfo sensorRegisterButton(int session_id, int button_id)
    {
        enterMethod();
        XSensorInfo info = null;
        try
        {
            info = mSimulationServer.sensorRegisterButton(session_id, button_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return info;
    }

    @Override
    public void sensorSwitchLightOn(int session_id, int port_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.sensorSwitchLightOn(session_id, port_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void sensorSwitchLightOff(int session_id, int port_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.sensorSwitchLightOff(session_id, port_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public float[] sensorGetValues(int session_id)
    {
        enterMethod();
        float[] values = null;
        try
        {
            values = mSimulationServer.sensorGetValues(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return values;
    }

    @Override
    public void displayRefresh(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.displayRefresh(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void displayClear(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.displayClear(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void displayClear(int session_id, int y_position)
    {
        enterMethod();
        try
        {
            mSimulationServer.displayClear(session_id, y_position);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void displayClear(int session_id, int x_position, int y_position, int count)
    {
        enterMethod();
        try
        {
            mSimulationServer.displayClear(session_id, x_position, y_position, count);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void displayDraw(int session_id, char ch, int x_position, int y_position)
    {
        enterMethod();
        try
        {
            mSimulationServer.displayDraw(session_id, ch, x_position, y_position);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void displayDraw(int session_id, int n, int x_position, int y_position)
    {
        enterMethod();
        try
        {
            mSimulationServer.displayDraw(session_id, n, x_position, y_position);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void displayDraw(int session_id, int n, int x_position, int y_position, int digits)
    {
        enterMethod();
        try
        {
            mSimulationServer.displayDraw(session_id, n, x_position, y_position, digits);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void displayDraw(int session_id, float fl, int x_position, int y_position)
    {
        enterMethod();
        try
        {
            mSimulationServer.displayDraw(session_id, fl, x_position, y_position);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void displayDraw(int session_id, String text, int x_position, int y_position)
    {
        enterMethod();
        try
        {
            mSimulationServer.displayDraw(session_id, text, x_position, y_position);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void displayDraw(int session_id, String text, int x_position, int y_position, boolean inverted)
    {
        enterMethod();
        try
        {
            mSimulationServer.displayDraw(session_id, text, x_position, y_position, inverted);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public int soundGetVolume(int session_id)
    {
        enterMethod();
        int volume = 0;
        try
        {
            volume = mSimulationServer.soundGetVolume(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return volume;
    }

    @Override
    public void soundSetVolume(int session_id, int volume)
    {
        enterMethod();
        try
        {
            mSimulationServer.soundSetVolume(session_id, volume);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public int soundBeep(int session_id)
    {
        enterMethod();
        int period = 0;
        try
        {
        	period = mSimulationServer.soundBeep(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
        return period;
    }

    @Override
    public int soundTwoBeeps(int session_id)
    {
        enterMethod();
        int period = 0;
        try
        {
        	period = mSimulationServer.soundTwoBeeps(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
        return period;
    }

    @Override
    public int soundBeepSequence(int session_id)
    {
        enterMethod();
        int period = 0;
        try
        {
        	period = mSimulationServer.soundBeepSequence(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
        return period;
    }

    @Override
    public int soundBeepSequenceUp(int session_id)
    {
        enterMethod();
        int period = 0;
        try
        {
        	period = mSimulationServer.soundBeepSequenceUp(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
        return period;
    }

    @Override
    public int soundBuzz(int session_id)
    {
        enterMethod();
        int period = 0;
        try
        {
        	period = mSimulationServer.soundBuzz(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
        return period;
    }

    @Override
    public void soundTone(int session_id, int frequency, int period)
    {
        enterMethod();
        try
        {
        	mSimulationServer.soundTone(session_id, frequency, period);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void soundTone(int session_id, int frequency, int period, int volume)
    {
        enterMethod();
        try
        {
            mSimulationServer.soundTone(session_id, frequency, period, volume);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void soundNote(int session_id, int instrument_id, int frequency, int period)
    {
        enterMethod();
        try
        {
            mSimulationServer.soundNote(session_id, instrument_id, frequency, period);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void consoleOpen(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.consoleOpen(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void consoleClose(int session_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.consoleClose(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void consolePrint(int session_id, String text)
    {
        enterMethod();
        try
        {
            mSimulationServer.consolePrint(session_id, text);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void consolePrintln(int session_id, String text)
    {
        enterMethod();
        try
        {
            mSimulationServer.consolePrintln(session_id, text);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public float motorGetMaxSpeed(int session_id, int port_id)
    {
        enterMethod();
        float speed = 0f;
        try
        {
            speed = mSimulationServer.motorGetMaxSpeed(session_id, port_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return speed;
    }

    @Override
    public void motorStartControl(int session_id, int port_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.motorStartControl(session_id, port_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void motorSetToFloatMode(int session_id, int port_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.motorSetToFloatMode(session_id, port_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public float[] motorGetPIDParameters(int session_id, int port_id)
    {
        enterMethod();
        float[] pid = null;
        try
        {
            pid = mSimulationServer.motorGetPIDParameters(session_id, port_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return pid;
    }

    @Override
    public float[] motorGetTachoCounts(int session_id)
    {
        enterMethod();
        float[] counts = null;
        try
        {
            counts = mSimulationServer.motorGetTachoCounts(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return counts;
    }

    @Override
    public void motorSetAllPowerData(int session_id, float[] power)
    {
        enterMethod();
        try
        {
            mSimulationServer.motorSetAllPowerData(session_id, power);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public float robotGetDefaultTrackWidth(int session_id)
    {
        enterMethod();
        float wheelbase = 0f;
        try
        {
            wheelbase = mSimulationServer.robotGetDefaultTrackWidth(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return wheelbase;
    }

    @Override
    public float robotGetDefaultWheelDiameter(int session_id)
    {
        enterMethod();
        float diameter = 0f;
        try
        {
            diameter = mSimulationServer.robotGetDefaultWheelDiameter(session_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();

        return diameter;
    }

    @Override
    public void robotSetTrackWidth(int session_id, float track_width)
    {
        enterMethod();
        try
        {
            mSimulationServer.robotSetTrackWidth(session_id, track_width);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void robotSetWheelDiameter(int session_id, float wheel_diameter)
    {
        enterMethod();
        try
        {
            mSimulationServer.robotSetWheelDiameter(session_id, wheel_diameter);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }

    @Override
    public void robotSetMotors(int session_id, int left_motor_port_id, int right_motor_port_id)
    {
        enterMethod();
        try
        {
            mSimulationServer.robotSetMotors(session_id, left_motor_port_id, right_motor_port_id);
        }
        catch (RemoteException e) {simulationAborted();}
        leaveMethod();
    }
}
